
from __future__ import unicode_literals

from django.db import migrations
from django.db.migrations.operations.base import Operation


class AddLanguage(Operation):

    reversible = True

    def __init__(self, name):
        self.name = name

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, *args, **kwargs):
        schema_editor.execute("""
            CREATE TEXT SEARCH CONFIGURATION
                public.%s ( COPY = pg_catalog.english );
        """ % self.name)
        schema_editor.execute("""
            ALTER TEXT SEARCH CONFIGURATION %s
                ALTER MAPPING FOR
                    asciiword,
                    asciihword,
                    hword_asciipart,
                    word,
                    hword,
                    hword_part
                WITH simple;
        """ % (self.name))

    def database_backwards(self, app_label, schema_editor, *args, **kwargs):
        schema_editor.execute(
            "DROP TEXT SEARCH CONFIGURATION %s" % self.name)

    def describe(self):
        return "Creates Text Search Configuration for {0}".format(self.name)


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        AddLanguage('polish'),
    ]




from unittest import TestCase
from unittest.mock import Mock

import pytest
from langid.langid import LanguageIdentifier

from lily.search.detector import LanguageDetector
from lily.base.events import EventFactory


class LanguageDetectorTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    def setUp(self):
        LanguageDetector.identifier = None
        self.detector = LanguageDetector()

    #
    # DETECT_LANGUAGE
    #
    def test_detect_language(self):

        self.mocker.patch.object(
            LanguageIdentifier, 'from_modelstring'
        ).return_value = Mock(
            rank=Mock(return_value=[
                ('en', 0.9),
                ('de', 0.4),
            ]))

        assert self.detector.detect('hello world') == [
            {'abbr': 'en', 'name': 'English'},
            {'abbr': 'de', 'name': 'German'},
        ]

    def test_detect_language__limit_by_probability(self):

        self.mocker.patch.object(
            LanguageIdentifier, 'from_modelstring'
        ).return_value = Mock(
            rank=Mock(return_value=[
                ('en', 0.9),
                ('fr', 0.6),
                ('de', 0.55),
            ]))
        self.mocker.patch.object(self.detector, 'DETECT_THRESHOLD_PROB', 0.56)

        assert self.detector.detect('hello world') == [
            {'abbr': 'en', 'name': 'English'},
            {'abbr': 'fr', 'name': 'French'},
        ]

    def test_detect_language__no_machted(self):

        self.mocker.patch.object(
            LanguageIdentifier, 'from_modelstring'
        ).return_value = Mock(rank=Mock(return_value=[]))

        with pytest.raises(EventFactory.BrokenRequest) as e:
            self.detector.detect('hi world')

        assert e.value.data == {
            '@event': 'UNSUPPORTED_LANGUAGE_DETECTED',
            '@type': 'error',
            'text': 'hi world',
        }

    def test_detect_language__limit_by_length(self):

        self.mocker.patch.object(
            LanguageIdentifier, 'from_modelstring'
        ).return_value = Mock(
            rank=Mock(return_value=[
                ('en', 0.9),
                ('fr', 0.6),
                ('de', 0.55),
                ('es', 0.1),
            ]))

        self.mocker.patch.object(self.detector, 'DETECT_THRESHOLD_LEN', 2)

        assert self.detector.detect('hello world') == [
            {'abbr': 'en', 'name': 'English'},
            {'abbr': 'fr', 'name': 'French'},
        ]

    def test_detect_language__no_results(self):

        self.mocker.patch.object(
            LanguageIdentifier, 'from_modelstring'
        ).return_value = Mock(rank=Mock(return_value=[]))

        with pytest.raises(EventFactory.BrokenRequest) as e:
            self.detector.detect('hello world')

        assert e.value.data == {
            '@event': 'UNSUPPORTED_LANGUAGE_DETECTED',
            '@type': 'error',
            'text': 'hello world',
        }

    #
    # DETECT_DB_CONF
    #
    def test_detect_db_conf__returns_simple_cases(self):

        conf = LanguageDetector().detect_db_conf('orange')

        assert conf == 'english'

    def test_detect_db_conf__returns_detect_results(self):

        # 1st call with portuguese
        conf = LanguageDetector().detect_db_conf(
            'Cómo estás, cuál es tu nombre')

        assert conf == 'spanish'

        # 2nd call with english
        conf = LanguageDetector().detect_db_conf(
            'hello man how are you doing')

        assert conf == 'english'

        # 3rd call with Esperanto
        conf = LanguageDetector().detect_db_conf('hallo wie gehts')

        assert conf == 'german'

    def test_detect_db_conf__min_detection_length(self):

        conf = LanguageDetector().detect_db_conf('wat!')

        assert conf == 'simple'




from unittest.mock import Mock, call
from enum import Enum

from django.test import TestCase
from django.db import models
import pytest
from lily_assistant.config import Config

from lily.base.models import EnumChoiceField
from lily.base import serializers, parsers
from lily.base.models import (
    array,
    const,
    JSONSchemaField,
    multischema,
    number,
    object,
    one_of,
    string,
    enum,
)
from lily.entrypoint.schema import (
    Schema,
    ArrayValue,
    SchemaRenderer,
    MissingSchemaMappingError,
    ForbiddenFieldError,
)


class Enumer(models.Model):

    nickname = EnumChoiceField(
        max_length=100,
        enum_name='human_nicknames',
        choices=[('Jack', 'Jack'), ('Abel', 'Abel')])

    class Brands(Enum):
        nice = 'nice'

        awesome = 'awesome'

    brand = EnumChoiceField(max_length=100, enum=Brands)

    class Meta:
        app_label = 'base'


class ForbiddenSerializer(serializers.Serializer):

    _type = 'forbidden'

    picks = serializers.ChoiceField(choices=('A', 'B'))


class Human(models.Model):

    name = models.CharField(max_length=100)

    age = models.IntegerField(null=True, blank=True)

    is_ready = models.BooleanField()

    pets = JSONSchemaField(schema=array(string()))

    class Meta:
        app_label = 'base'


class JSONSchemaSerializer(serializers.Serializer):

    _type = 'json'

    names = serializers.JSONSchemaField(schema=array(string()))

    users = serializers.SerializerMethodField()

    owners = serializers.SerializerMethodField()

    class Shopping(Enum):
        bread = 'bread'

        milk = 'milk'

    shoppings = serializers.JSONSchemaField(
        schema=array(
            object(type=enum(Shopping))))

    selected_shoppings = serializers.JSONSchemaField(
        schema=one_of(
            object(type=enum(Shopping, const=Shopping.bread.value)),
            object(type=enum(Shopping, const=Shopping.milk.value))))

    def get_users(self, instance) -> [object(name=const('YO'), age=number())]:  # noqa
        return instance['users']

    def get_owners(self, instance) -> array(object(name=string(), age=number())):  # noqa
        return instance['owners']


class JSONSchemaMultiSchemaModel(models.Model):

    entity = JSONSchemaField(schema=multischema(
        {
            'PERSON': object(
                type=const('PERSON'),
                name=string(),
                required=['type', 'name']),
            'ANIMAL': object(
                type=const('ANIMAL'),
                age=number(),
                required=['type', 'age'])
        },
        by_field='type'))

    class Meta:
        app_label = 'base'


class JSONSchemaMultiSchemaSerializer(serializers.ModelSerializer):

    _type = 'json_multischema'

    class Meta:
        model = JSONSchemaMultiSchemaModel
        fields = ('entity',)


class HumanQueryParser(parsers.Parser):

    name = parsers.CharField(max_length=123, required=False)

    age = parsers.IntegerField(min_value=18)


class HumanBodyParser(parsers.Parser):

    name = parsers.CharField(max_length=123, required=False)

    age = parsers.IntegerField(min_value=18)


class HumanSerializer(serializers.Serializer):

    _type = 'human'

    name = serializers.CharField(max_length=123, required=False)

    age = serializers.IntegerField(min_value=18)

    place = serializers.EnumChoiceField(choices=[0, 1, 2], enum_name='place')


class CycleJSONSerializer(serializers.Serializer):

    _type = 'cycle'

    local_id = serializers.IntegerField(allow_null=True)

    stage_id = serializers.IntegerField(allow_null=True)

    entity = serializers.JSONField()


class CyclesJSONSerializer(serializers.Serializer):

    _type = 'cycles'

    card_cycle = serializers.SerializerMethodField()

    path_cycle = serializers.SerializerMethodField()

    def get_card_cycle(self, instance) -> [CycleJSONSerializer]:
        return []

    def get_path_cycle(self, instance) -> [CycleJSONSerializer]:
        return []


class HumanModelSerializer(serializers.ModelSerializer):

    _type = 'human'

    is_underaged = serializers.SerializerMethodField()

    class Meta:
        model = Human
        fields = (
            'name', 'age', 'is_ready', 'is_underaged', 'pets')

    def get_is_underaged(self, instance) -> bool:
        return instance.age > 18


class EnumerModelSerializer(serializers.ModelSerializer):

    _type = 'enumer'

    class Meta:
        model = Enumer
        fields = ('nickname', 'brand')


class ProviderSerializer(serializers.Serializer):

    _type = 'provider'

    name = serializers.CharField(default='Johnny')


class FoodSerializer(serializers.Serializer):

    _type = 'food'

    provider = ProviderSerializer()

    amount = serializers.IntegerField()


class PartySerializer(serializers.Serializer):

    _type = 'party'

    guests = HumanSerializer(many=True)

    host = HumanModelSerializer()

    food = FoodSerializer()


class SchemaRendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    #
    # Body Parser
    #
    def test_person_body_parser(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(
            HumanBodyParser,
            parser_type=parsers.ParserTypeEnum.BODY.value
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'integer',
                        'minimum': 18,
                    },
                    'name': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
        }

    def test_person_query_parser(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(
            HumanQueryParser,
            parser_type=parsers.ParserTypeEnum.QUERY.value
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'integer',
                        'minimum': 18,
                    },
                    'name': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
        }

    #
    # Serializer
    #
    def test_person_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(HumanSerializer).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': ['age', 'place'],
                'entity_type': 'human',
                'properties': {
                    'age': {
                        'type': 'integer',
                        'minimum': 18,
                    },
                    'name': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                    'place': {
                        'enum': [0, 1, 2],
                        'enum_name': 'place',
                        'type': 'integer',
                    },
                },
            },
        }

    def test_object_json_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(CycleJSONSerializer).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'entity_type': 'cycle',
                'properties': {
                    'local_id': {'type': 'integer'},
                    'stage_id': {'type': 'integer'},
                    'entity': {'type': 'any'},
                },
                'required': ['local_id', 'stage_id', 'entity'],
            },
        }

    def test_array_json_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(CyclesJSONSerializer).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'entity_type': 'cycles',
                'properties': {
                    'card_cycle': {
                        'type': 'array',
                        'items': {
                            'entity_type': 'cycle',
                            'type': 'object',
                            'properties': {
                                'entity': {'type': 'any'},
                                'local_id': {'type': 'integer'},
                                'stage_id': {'type': 'integer'},
                            },
                            'required': ['local_id', 'stage_id', 'entity'],
                        },
                    },
                    'path_cycle': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'entity_type': 'cycle',
                            'properties': {
                                'entity': {'type': 'any'},
                                'local_id': {'type': 'integer'},
                                'stage_id': {'type': 'integer'},
                            },
                            'required': ['local_id', 'stage_id', 'entity'],
                        },
                    },
                },
                'required': ['card_cycle', 'path_cycle'],
            },
        }

    def test_method_json_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(JSONSchemaSerializer).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'entity_type': 'json',
                'properties': {
                    'names': {
                        'type': 'array',
                        'items': {'type': 'string'},
                    },
                    'owners': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'age': {'type': 'number'},
                                'name': {'type': 'string'},
                            },
                        },
                    },
                    'users': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'age': {'type': 'number'},
                                'name': {
                                    'type': 'string',
                                    'pattern': '^YO$',
                                    'const': 'YO',
                                },
                            },
                        },
                    },
                    'selected_shoppings': {
                        'by_field': None,
                        'oneOf': [
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'const_value': 'bread',
                                        'enum': ['bread', 'milk'],
                                        'enum_name': 'Shopping',
                                        'type': 'string'
                                    }
                                },
                            },
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'const_value': 'milk',
                                        'enum': ['bread', 'milk'],
                                        'enum_name': 'Shopping',
                                        'type': 'string'
                                    }
                                },
                            },
                        ],
                    },
                    'shoppings': {
                        'type': 'array',
                        'items': {
                            'properties': {
                                'type': {
                                    'enum': ['bread', 'milk'],
                                    'enum_name': 'Shopping',
                                    'type': 'string'
                                }
                            },
                            'type': 'object'
                        },
                    },

                },
                'required': [
                    'names',
                    'users',
                    'owners',
                    'shoppings',
                    'selected_shoppings',
                ],
            },
        }

    def test_method_json_multischema_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(
            JSONSchemaMultiSchemaSerializer
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': ['entity'],
                'properties': {
                    'entity': {
                        'oneOf': [
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'type': 'string',
                                        'const': 'PERSON',
                                        'pattern': '^PERSON$'
                                    },
                                    'name': {
                                        'type': 'string'
                                    }
                                },
                                'required': ['type', 'name']
                            },
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'type': 'string',
                                        'const': 'ANIMAL',
                                        'pattern': '^ANIMAL$'
                                    },
                                    'age': {
                                        'type': 'number'
                                    }
                                },
                                'required': ['type', 'age']
                            }
                        ]
                    }
                },
                'entity_type': 'json_multischema',
            }
        }

    #
    # Model Serializer
    #
    def test_person_model_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(
            HumanModelSerializer
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'entity_type': 'human',
                'properties': {
                    'name': {
                        'type': 'string',
                        'maxLength': 100,
                    },
                    'age': {
                        'minimum': -2147483648,
                        'maximum': 2147483647,
                        'type': 'integer',
                    },
                    'is_underaged': {
                        'type': 'boolean',
                    },
                    'is_ready': {
                        'type': 'boolean',
                    },
                    'pets': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    },
                },
                'required': ['name', 'is_ready', 'is_underaged', 'pets'],
            },
        }

    def test_enumer_model_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        assert SchemaRenderer(
            EnumerModelSerializer
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'entity_type': 'enumer',
                'properties': {
                    'brand': {
                        'enum': ['awesome', 'nice'],
                        'enum_name': 'Brands',
                        'type': 'string'
                    },
                    'nickname': {
                        'enum': ['Abel', 'Jack'],
                        'enum_name': 'human_nicknames',
                        'type': 'string',
                    },
                },
                'required': ['nickname', 'brand'],
                'type': 'object',
            },
        }

    def test_forbidden_serializer(self):

        with pytest.raises(ForbiddenFieldError) as e:
            SchemaRenderer(ForbiddenSerializer).render().serialize()

        assert e.value.args == (
            'picks uses ChoiceField which is forbidden since it leads to '
            'ambiguous client side Enums. Please use `EnumChoiceField` '
            'instead',)

    #
    # Method Derived Serializer Fields
    #
    def test_simple_fields(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        class SimpleFieldsSerializer(serializers.Serializer):

            number = serializers.IntegerField()

            is_correct = serializers.BooleanField()

            price = serializers.FloatField()

            choice = serializers.EnumChoiceField(
                enum_name='choice',
                choices=[('AA', 'aaa'), ('BB', 'bbb')])

            name = serializers.CharField(max_length=11)

            data = serializers.DictField(required=False)

            email = serializers.EmailField()

            uri = serializers.URLField(required=False)

            amount = serializers.DecimalField(
                required=False, max_digits=5, decimal_places=3)

            json = serializers.JSONField()

            date_of_birth = serializers.DateField()

            created_at = serializers.DateTimeField()

        assert SchemaRenderer(
            SimpleFieldsSerializer
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': [
                    'number', 'is_correct', 'price', 'choice', 'name',
                    'email', 'json', 'date_of_birth', 'created_at'],
                'properties': {
                    'amount': {
                        'type': 'string',
                    },
                    'number': {
                        'type': 'integer',
                    },
                    'is_correct': {
                        'type': 'boolean',
                    },
                    'price': {
                        'type': 'number',
                    },
                    'choice': {
                        'type': 'string',
                        'enum': ['AA', 'BB'],
                        'enum_name': 'choice',
                    },
                    'name': {
                        'maxLength': 11,
                        'type': 'string',
                    },
                    'data': {
                        'type': 'object',
                    },
                    'email': {
                        'type': 'string',
                        'format': 'email',
                    },
                    'uri': {
                        'type': 'string',
                        'format': 'uri',
                    },
                    'json': {
                        'type': 'any',
                    },
                    'date_of_birth': {
                        'type': 'string',
                        'format': 'date',
                    },
                    'created_at': {
                        'type': 'string',
                        'format': 'date-time',
                    }
                },
            }
        }

    def test_list_field(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        class Serializer(serializers.Serializer):

            _type = 'complex'

            numbers = serializers.ListField(
                child=serializers.IntegerField())

            person = serializers.ListField(
                child=HumanSerializer())

            # -- optional list fields
            ids = serializers.ListField(
                child=serializers.IntegerField(), required=False)

            providers = serializers.ListField(
                child=ProviderSerializer(), required=False)

        assert SchemaRenderer(Serializer).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'entity_type': 'complex',
                'required': ['numbers', 'person'],
                'properties': {
                    'person': {
                        'type': 'array',
                        'items': {
                            'required': ['age', 'place'],
                            'type': 'object',
                            'entity_type': 'human',
                            'properties': {
                                'name': {
                                    'maxLength': 123,
                                    'type': 'string'
                                },
                                'age': {
                                    'type': 'integer',
                                    'minimum': 18
                                },
                                'place': {
                                    'enum': [0, 1, 2],
                                    'enum_name': 'place',
                                    'type': 'integer',
                                },
                            },
                        },
                    },
                    'numbers': {
                        'type': 'array',
                        'items': {
                            'type': 'integer'
                        },
                    },
                    'ids': {
                        'type': 'array',
                        'items': {
                            'type': 'integer'
                        },
                    },
                    'providers': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'entity_type': 'provider',
                            'properties': {
                                'name': {
                                    'type': 'string',
                                }
                            },
                            'required': [],
                        },
                    },
                },
            },
        }

    #
    # Method Derived Serializer Fields
    #
    def test_method_derived_fields(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        class MethodDerivedFieldsSerializer(serializers.Serializer):

            _type = 'method'

            number = serializers.SerializerMethodField()

            ingredients = serializers.SerializerMethodField()

            names = serializers.SerializerMethodField()

            hosts = serializers.SerializerMethodField()

            owner = serializers.SerializerMethodField()

            def get_number(self, *args, **kwargs) -> float:
                return 1.45

            def get_ingredients(self, *args, **kwargs) -> [int]:
                return [1, 34]

            def get_names(self, *args, **kwargs) -> [str]:
                return ['Jake', 'John']

            def get_hosts(self, *args, **kwargs) -> [HumanSerializer]:
                return [{'name': 'John', 'age': 24}]

            def get_owner(self, *args, **kwargs) -> HumanSerializer:
                return {'name': 'John', 'age': 24}

        assert SchemaRenderer(
            MethodDerivedFieldsSerializer
        ).render().serialize() == {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': [
                    'number', 'ingredients', 'names', 'hosts', 'owner'],
                'entity_type': 'method',
                'properties': {
                    'number': {
                        'type': 'number',
                    },
                    'ingredients': {
                        'type': 'array',
                        'items': {
                            'type': 'number',
                        },
                    },
                    'names': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    },
                    'hosts': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'required': ['age', 'place'],
                            'entity_type': 'human',
                            'properties': {
                                'name': {
                                    'type': 'string',
                                    'maxLength': 123,
                                },
                                'age': {
                                    'type': 'integer',
                                    'minimum': 18,
                                },
                                'place': {
                                    'enum': [0, 1, 2],
                                    'enum_name': 'place',
                                    'type': 'integer',
                                },
                            },
                        },
                    },
                    'owner': {
                        'type': 'object',
                        'entity_type': 'human',
                        'required': ['age', 'place'],
                        'properties': {
                            'name': {
                                'type': 'string',
                                'maxLength': 123,
                            },
                            'age': {
                                'type': 'integer',
                                'minimum': 18,
                            },
                            'place': {
                                'enum': [0, 1, 2],
                                'enum_name': 'place',
                                'type': 'integer',
                            },
                        },
                    },
                },
            },
        }

    #
    # Nested Serializer
    #
    def test_nested_party_serializer(self):

        self.mocker.patch.object(
            Schema, 'get_repository_uri'
        ).return_value = 'http://hi.there#123'

        expected = {
            'uri': 'http://hi.there#123',
            'schema': {
                'type': 'object',
                'required': ['guests', 'host', 'food'],
                'entity_type': 'party',
                'properties': {
                    'host': {
                        'type': 'object',
                        'required': [
                            'name',
                            'is_ready',
                            'is_underaged',
                            'pets',
                        ],
                        'entity_type': 'human',
                        'properties': {
                            'name': {
                                'type': 'string',
                                'maxLength': 100,
                            },
                            'age': {
                                'type': 'integer',
                                'minimum': -2147483648,
                                'maximum': 2147483647
                            },
                            'is_underaged': {
                                'type': 'boolean',
                            },
                            'is_ready': {
                                'type': 'boolean',
                            },
                            'pets': {
                                'type': 'array',
                                'items': {
                                    'type': 'string',
                                },
                            }
                        },
                    },
                    'food': {
                        'type': 'object',
                        'entity_type': 'food',
                        'required': ['provider', 'amount'],
                        'properties': {
                            'provider': {
                                'type': 'object',
                                'entity_type': 'provider',
                                'properties': {
                                    'name': {
                                        'type': 'string',
                                    },
                                },
                                'required': [],
                            },
                            'amount': {
                                'type': 'integer',
                            },
                        },
                    },
                    'guests': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'entity_type': 'human',
                            'required': ['age', 'place'],
                            'properties': {
                                'name': {
                                    'type': 'string',
                                    'maxLength': 123,
                                },
                                'age': {
                                    'type': 'integer',
                                    'minimum': 18,
                                },
                                'place': {
                                    'enum': [0, 1, 2],
                                    'enum_name': 'place',
                                    'type': 'integer',
                                },
                            },
                        },
                    },
                },
            },
        }

        assert SchemaRenderer(
            PartySerializer).render().serialize() == expected

    def test_missing_mapping(self):

        class UnknownField:
            pass

        class Serializer:

            def get_fields(self):
                return {'name': UnknownField()}

        with pytest.raises(MissingSchemaMappingError):
            SchemaRenderer(Serializer).render().serialize()

    #
    # test_get_repository_uri
    #
    def test_get_repository_uri__bitbucket(self):

        class MockConfig:

            repository = 'http://bitbucket.com'

            last_commit_hash = '1234'

            @classmethod
            def get_project_path(cls):
                return '/home/projects/lily'

        self.mocker.patch(
            'lily.entrypoint.schema.Config', MockConfig)

        schema = SchemaRenderer(HumanSerializer).render()
        schema.meta = {
            'path': '/some/path',
            'first_line': 11,
        }

        assert (
            schema.get_repository_uri() ==
            'http://bitbucket.com/src/1234/some/path/#lines-11')

    def test_get_repository_uri__github(self):

        class MockConfig:

            repository = 'http://github.com/what'

            last_commit_hash = '1234'

            @classmethod
            def get_project_path(cls):
                return '/home/projects/lily'

        self.mocker.patch(
            'lily.entrypoint.schema.Config', MockConfig)

        schema = SchemaRenderer(HumanSerializer).render()
        schema.meta = {
            'path': '/some/path',
            'first_line': 11,
        }

        assert (
            schema.get_repository_uri() ==
            'http://github.com/what/blob/1234/some/path/#L11')

    #
    # get_meta
    #
    def test_get_meta(self):

        self.mocker.patch.object(
            Config, 'get_project_path').return_value = '/home/projects/lily'
        getfile = Mock(return_value='/home/projects/lily/a/views.py')
        getsourcelines = Mock(return_value=[None, 11])
        self.mocker.patch(
            'lily.entrypoint.schema.inspect',
            Mock(getfile=getfile, getsourcelines=getsourcelines))

        renderer = SchemaRenderer(HumanSerializer)

        assert renderer.get_meta(HumanSerializer) == {
            'first_line': 11,
            'path': '/a/views.py',
        }
        assert getfile.call_args_list == [
            call(HumanSerializer), call(HumanSerializer)]
        assert getsourcelines.call_args_list == [
            call(HumanSerializer), call(HumanSerializer)]


def get_schema(raw_schema):

    schema = SchemaRenderer(HumanSerializer).render()
    schema.schema = raw_schema

    return schema


@pytest.mark.parametrize(
    'raw_schema, expected',
    [
        # -- empty to empty
        (
            {
                'type': 'object',
                'properties': {},
                'required': [],
            },
            {'schema': {}},
        ),

        # -- single value schema
        (
            {
                'type': 'object',
                'properties': {
                    'name': 'string',
                },
                'required': ['name'],
            },
            {
                'schema': {
                    'type': 'object',
                    'properties': {
                        'name': 'string',
                    },
                    'required': ['name'],
                },
            }
        ),

        #
        # Nested Schemas
        #
        (
            {
                'type': 'object',
                'required': ['owner'],
                'properties': {
                    'owner': get_schema({
                        'type': 'object',
                        'properties': {
                            'name': 'string',
                            'ingredient': get_schema({
                                'type': 'object',
                                'required': ['is_ready', 'name'],
                                'properties': {
                                    'name': 'string',
                                    'is_ready': 'boolean',
                                    'creator': get_schema({
                                        'type': 'object',
                                        'required': [],
                                        'properties': {
                                            'age': 'number',
                                        },
                                    }),
                                },
                            }),
                        },
                        'required': ['name'],
                    }),
                },
            },
            {
                'schema': {
                    'type': 'object',
                    'required': ['owner'],
                    'properties': {
                        'owner': {
                            'type': 'object',
                            'properties': {
                                'name': 'string',
                                'ingredient': {
                                    'type': 'object',
                                    'required': ['is_ready', 'name'],
                                    'properties': {
                                        'name': 'string',
                                        'is_ready': 'boolean',
                                        'creator': {
                                            'type': 'object',
                                            'required': [],
                                            'properties': {
                                                'age': 'number',
                                            },
                                        },
                                    },
                                },
                            },
                            'required': ['name'],
                        },
                    },
                },
            }
        ),

        #
        # ArrayValue
        #

        # -- Schema - ArrayValue
        (
            {
                'type': 'object',
                'required': ['owner'],
                'properties': {
                    'owner': ArrayValue(get_schema({
                        'type': 'object',
                        'required': [],
                        'properties': {
                            'name': 'string',
                            'age': 'number',
                        },
                    })),
                },
            },
            {
                'schema': {
                    'type': 'object',
                    'required': ['owner'],
                    'properties': {
                        'owner': {
                            'type': 'array',
                            'items': {
                                'type': 'object',
                                'required': [],
                                'properties': {
                                    'name': 'string',
                                    'age': 'number',
                                },
                            },
                        },
                    },
                },
            }
        ),

        # -- simple type - ArrayValue
        (
            {
                'type': 'object',
                'required': ['choices'],
                'properties': {
                    'choices': ArrayValue('string'),
                }
            },
            {
                'schema': {
                    'type': 'object',
                    'required': ['choices'],
                    'properties': {
                        'choices': {
                            'type': 'array',
                            'items': 'string',
                        },
                    }
                },
            }
        ),
    ])
def test_serialize(raw_schema, expected, mocker):

    mocker.patch.object(
        Schema, 'get_repository_uri').return_value = 'uri'

    schema = SchemaRenderer(HumanSerializer).render()
    schema.schema = raw_schema

    serialized = schema.serialize()
    del serialized['uri']
    assert serialized == expected




import os

from django.test import TestCase
import pytest

from lily.base.source import Source, SourceSerializer


class SourceTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initifixtures(self, mocker):
        self.mocker = mocker

    def setUp(self):
        class ConfigMock:

            @classmethod
            def get_project_path(cls):
                return os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    '..',
                    '..')

        self.mocker.patch('base.source.Config', ConfigMock)

    def test_constructor(self):

        def fn():
            a = 1
            b = a
            return b

        source = Source(fn)

        assert source.filepath == '/tests/test_base/test_source.py'
        assert source.start_line == 30
        assert source.end_line == 33


class SourceSerializerTestCase(TestCase):

    def test_serialization(self):

        def fn():
            a = 1
            return a

        source = Source(fn)

        assert SourceSerializer(source).data == {
            '@type': 'source',
            'filepath': '/tests/test_base/test_source.py',
            'start_line': 46,
            'end_line': 48,
        }




import re

import inflect

from .events import EventFactory


INFLECT_ENGINE = inflect.engine()


IRREGULAR_FORMS = {
    'break': 'broke',
    'broadcast': 'broadcast',
    'buy': 'bought',
    'come': 'came',
    'cost': 'cost',
    'cut': 'cut',
    'eat': 'ate',
    'feel': 'felt',
    'get': 'got',
    'go': 'went',
    'hear': 'heard',
    'hit': 'hit',
    'hurt': 'hurt',
    'is': 'was',
    'let': 'let',
    'lose': 'lost',
    'make': 'made',
    'pay': 'paid',
    'put': 'put',
    'quit': 'quit',
    'read': 'read',
    'reset': 'reset',
    'reset': 'reset',
    'see': 'saw',
    'send': 'sent',
    'sing': 'sang',
    'sit': 'sat',
    'sleep': 'slept',
    'spread': 'spread',
    'teach': 'taught',
    'tell': 'told',
    'understand': 'understood',
}


EXCEPTION_FORMS = {
    'refer': 'referred',
}


VOWELS = set(['a', 'e', 'i', 'o', 'u', 'y'])


END_HUSHERS = set([
    'ch',
    'ck',
    'h',
    'k',
    'lk',
    'lp',
    'nt',
    'sh',
    'st',
    'wn',
    'x',
])


def to_plural(noun):
    """Forgiving plural form transformer.

    If the `noun` is already in the plural form no transformation will be
    applied.

    """
    if not INFLECT_ENGINE.singular_noun(noun):
        return INFLECT_ENGINE.plural(noun)

    return noun


def to_past(verb):

    verb = verb.lower()
    if verb in IRREGULAR_FORMS:
        return IRREGULAR_FORMS[verb]

    elif verb in EXCEPTION_FORMS:
        return EXCEPTION_FORMS[verb]

    # -- last letter is `e`
    elif verb[-1] == 'e':
        return verb + 'd'

    # -- last letter is `y`
    elif verb[-1] == 'y':
        if verb[-2] in VOWELS:
            return verb + 'ed'

        else:
            return verb[:-1] + 'ied'

    elif needs_double_consonant(verb):
        return verb + verb[-1] + 'ed'

    else:
        return verb + 'ed'


def needs_double_consonant(verb):

    vowels_count = len([letter for letter in verb if letter in VOWELS])

    return not (
        vowels_count > 1 or         # -- is monosyllabic word
        verb[-1] in END_HUSHERS or   # -- last letter is husher
        verb[-2:] in END_HUSHERS or  # -- last two letters are hushers
        verb[-1] == verb[-2]         # -- last two letters already repeated
    )


class BaseVerb:

    verb = NotImplemented

    finalizers = NotImplemented

    def __init__(self, noun):

        # -- verb
        self.verb = self.transform(self.verb)

        # -- past_verb
        self.past_verb = self.transform(to_past(self.verb))

        # -- noun
        if isinstance(noun, str):
            self.noun = self.transform(noun)

        else:
            self.noun = self.transform(noun._meta.model_name)

    def render_command_name(self):

        return '{verb}_{noun}'.format(
            verb=self.verb, noun=self.noun)

    def render_event_name(self, request=None, e=None):

        if request and e:
            if not isinstance(e, self.finalizers):
                raise EventFactory.BrokenRequest(
                    event=(
                        'INVALID_FINALIZER_USED_FOR_SPECIFIC_COMMAND_'
                        'DETECTED'),
                    context=request)

            return '{noun}_{past_verb}'.format(
                past_verb=self.transform(e.verb or e.__class__.__name__),
                noun=self.noun)

        else:
            return '{noun}_{past_verb}'.format(
                past_verb=self.past_verb,
                noun=self.noun)

    def transform(self, name):
        name = name.upper()
        name = re.sub(r'\s+', '_', name)
        name = re.sub(r'\_{2,}', '_', name)

        return name

    @property
    def after(self):
        return Conjunction(effect=self)


class BaseBulkVerb(BaseVerb):

    def __init__(self, noun):

        if not isinstance(noun, str):
            noun = noun._meta.model_name

        super(BaseBulkVerb, self).__init__(to_plural(noun))


class Conjunction:

    def __init__(self, effect):
        self.effect = effect

        # -- Generic Execute
        self.Execute = self.phrase(Execute, effect)

        # -- CRUD
        self.Create = self.phrase(Create, effect)
        self.Read = self.phrase(Read, effect)
        self.Update = self.phrase(Update, effect)
        self.Delete = self.phrase(Delete, effect)

        # -- BULK CRUD
        self.BulkCreate = self.phrase(BulkCreate, effect)
        self.BulkRead = self.phrase(BulkRead, effect)
        self.BulkUpdate = self.phrase(BulkUpdate, effect)
        self.BulkDelete = self.phrase(BulkDelete, effect)

        # -- CONDITIONAL CRUD
        self.CreateOrUpdate = self.phrase(CreateOrUpdate, effect)
        self.CreateOrRead = self.phrase(CreateOrRead, effect)

    def phrase(self, cause_cls, effect):

        class Phrase:

            effect = None

            def __init__(other, *args, **kwargs):  # noqa
                self.cause = cause_cls(*args, **kwargs)

            def render_command_name(other):  # noqa
                return '{effect}_AFTER_{cause}'.format(
                    effect=self.effect.render_command_name(),
                    cause=self.cause.render_command_name())

            def render_event_name(other, request, e):  # noqa
                return '{effect}_AFTER_{cause}'.format(
                    effect=self.effect.render_event_name(request, e),
                    cause=self.cause.render_event_name())

        Phrase.effect = effect
        return Phrase


class ConstantName(BaseVerb):

    def __init__(self, command_name):
        self.command_name = command_name

    def render_command_name(self):
        return self.transform(self.command_name)

    def render_event_name(self, request, e):

        return self.transform(e.event)


class Execute(BaseVerb):

    verb = NotImplemented

    finalizers = (EventFactory.Executed,)

    def __init__(self, verb, noun):
        self.verb = verb

        super(Execute, self).__init__(noun)

    def render_event_name(self, request=None, e=None):

        if request and e:
            if not isinstance(e, self.finalizers):
                raise EventFactory.BrokenRequest(
                    event=(
                        'INVALID_FINALIZER_USED_FOR_SPECIFIC_COMMAND_'
                        'DETECTED'),
                    context=request)

        return '{noun}_{past_verb}'.format(
            past_verb=self.past_verb,
            noun=self.noun)


#
# CRUD
#
class Create(BaseVerb):

    finalizers = (EventFactory.Created,)

    verb = 'create'


class Read(BaseVerb):

    finalizers = (EventFactory.Read,)

    verb = 'read'


class Update(BaseVerb):

    finalizers = (EventFactory.Updated,)

    verb = 'update'


class Delete(BaseVerb):

    finalizers = (EventFactory.Deleted,)

    verb = 'delete'


#
# BULK CRUD
#
class BulkCreate(BaseBulkVerb):

    finalizers = (EventFactory.BulkCreated,)

    verb = 'bulk_create'


class BulkRead(BaseBulkVerb):

    finalizers = (EventFactory.BulkRead,)

    verb = 'bulk_read'


class BulkUpdate(BaseBulkVerb):

    finalizers = (EventFactory.BulkUpdated,)

    verb = 'bulk_update'


class BulkDelete(BaseBulkVerb):

    finalizers = (EventFactory.BulkDeleted,)

    verb = 'bulk_delete'


#
# CONDITIONAL CRUD
#
class CreateOrUpdate(BaseVerb):

    finalizers = (EventFactory.Created, EventFactory.Updated)

    verb = 'create_or_update'


class CreateOrRead(BaseVerb):

    finalizers = (EventFactory.Created, EventFactory.Read)

    verb = 'read_or_create'




import re
import random

import pytest

from search.latex import transformer
from tests.test_search import LATEX_CASES


def _normalize(x):
    x = re.sub(r'\s+', ' ', x)
    return x.strip()


def _normalize_desc(x):
    x = re.sub(r'\s+', '_', x)
    return re.sub(r'\W+', '', x)


def _select_cases(cases):

    def _select_case(c):
        priority_test_run = (
            any(map(lambda c: c.get('run_test', False), cases)))
        return (
            not priority_test_run or
            (priority_test_run and c.get('run_test', False)))

    return [c for c in cases if _select_case(c)]


@pytest.mark.parametrize(
    "text, english_expected, polish_expected",
    [[
        c['text'],
        c['english_expected'],
        c['polish_expected']]
     for c in _select_cases(LATEX_CASES)],
    ids=[_normalize_desc(c['description'])
         for c in _select_cases(LATEX_CASES)])
def test_transform(text, english_expected, polish_expected, mocker):

    mocker.patch('search.latex.transformer.MAX_ITERATIONS_COUNT', 3)

    assert (
        _normalize(transformer.transform(text, 'english')) ==
        _normalize(english_expected))

    assert (
        _normalize(transformer.transform(text, 'polish')) ==
        _normalize(polish_expected))

    # here we can randomize all other languages to make sure that they
    # always default to english is not supported
    other_languages = [
        'danish', 'dutch', 'finnish', 'french', 'german', 'hungarian',
        'italian', 'norwegian', 'portuguese', 'romanian', 'russian',
        'simple', 'spanish', 'swedish', 'turkish',
    ]
    random_language = random.choice(other_languages)
    assert (
        _normalize(transformer.transform(text, random_language)) ==
        _normalize(english_expected))




from datetime import datetime
import string
from unittest.mock import call

from itsdangerous import SignatureExpired, URLSafeTimedSerializer
import pytest

from lily.base import signature
from lily.base.events import EventFactory


#
# sign_payload
#
def test_sign_payload__calls_dumps_correctly(mocker):

    dumps_mock = mocker.patch.object(URLSafeTimedSerializer, 'dumps')

    signature.sign_payload(
        email='test@whats.com',
        payload='WHATEVER',
        secret='personal_secret',
        salt='salt_me')

    assert dumps_mock.call_args_list == [
        call(['test@whats.com', 'WHATEVER'], salt='salt_me')]


def test_sign_payload__returns_correct_code(mocker):

    mocker.patch.object(
        URLSafeTimedSerializer, 'dumps').return_value = 'g67g6f7g'

    encoded_payload = signature.sign_payload(
        email='test@whats.com',
        payload='WAT',
        secret='personal_secret',
        salt='salt_me')

    assert encoded_payload == 'g67g6f7g'


#
# verify_payload
#
def test_verify_payload__make_the_right_calls(mocker):

    loads_mock = mocker.patch.object(URLSafeTimedSerializer, 'loads')
    loads_mock.return_value = ('hi@there', 'HI')

    payload = signature.verify_payload(
        encoded_payload='rubishtokenwhereareyou',
        secret='my_secret',
        salt='salt_me',
        signer_email='hi@there',
        max_age=24)

    assert payload == 'HI'
    assert loads_mock.call_count == 1


def test_verify_payload__different_secrets_for_encoding_and_decoding():

    code = signature.sign_payload(
        'why@gmail.com', 'NO!', 'secret123', 'salt')

    with pytest.raises(EventFactory.BrokenRequest) as e:
        assert signature.verify_payload(
            code, 'my_secret', 'salt', 'hi@there', 120)

    assert e.value.event == 'PAYLOAD_VERIFIED_AS_BROKEN'


def test_verify_payload__email_mismatch(mocker):

    code = signature.sign_payload(
        'why@gmail.com', 'NO!', 'secret123', 'salt')

    with pytest.raises(EventFactory.BrokenRequest) as e:
        signature.verify_payload(
            code, 'secret123', 'salt', 'hello@there', 120)

    assert e.value.event == 'PAYLOAD_VERIFIED_AS_BROKEN_MISMATCHING_EMAILS'


def test_verify_payload__recognizes_expired_token(mocker):

    mocker.patch.object(
        URLSafeTimedSerializer,
        'loads'
    ).side_effect = SignatureExpired(
        'error occured',
        date_signed=datetime(2013, 1, 15, 6, 48))

    with pytest.raises(EventFactory.BrokenRequest) as e:
        signature.verify_payload(
            'what.ever', 'personal_secret', 'salt', 'test@whats.com', 24)

    assert e.value.event == 'PAYLOAD_VERIFIED_AS_EXPIRED'


#
# create_secret
#
def test_create_secret__creates_unique_secrets():

    secrets = [signature.create_secret() for i in range(1000)]

    assert len(secrets) == len(set(secrets))


def test_create_secret__safe_secret():

    def assert_is_secret_safe(secret):
        assert len(secret) == 64

        assert len(set(string.ascii_uppercase) & set(secret)) > 0
        assert len(set(string.ascii_lowercase) & set(secret)) > 0
        assert len(set(string.digits) & set(secret)) > 0
        assert len(set(string.punctuation) & set(secret)) > 0

        # uniqueness of characters
        assert len(set(secret)) > 30

    for i in range(1000):
        assert_is_secret_safe(signature.create_secret())




from lily.base.events import EventFactory


class BaseAuthorizer(EventFactory):
    """Minimal Authorizer Class."""

    def __init__(self, access_list):
        self.access_list = access_list

    def authorize(self, request):
        try:
            return {
                'user_id': request.META['HTTP_X_CS_USER_ID'],
                'account_type': request.META['HTTP_X_CS_ACCOUNT_TYPE'],
            }

        except KeyError:
            raise self.AccessDenied('ACCESS_DENIED', context=request)

    def log(self, authorize_data):
        return authorize_data




import json

from .interface import Interface
from .signature import Signature
from .utils import to_camelcase
from lily.base.utils import normalize_indentation


class Command:

    def __init__(self, name, conf):
        self.name = name
        self.conf = conf

        # -- access attributes
        self.method = self.conf['method'].lower()
        self.camel_name = to_camelcase(self.name, first_lower=True)

        # -- META
        self.domain_id = self.conf['meta']['domain']['id']
        self.domain_name = self.conf['meta']['domain']['name']
        self.title = self.conf['meta']['title']

        # -- ACCESS
        self.is_private = self.conf['access']['is_private']
        self.authorization_required = (
            self.conf['access']['access_list'] is not None)
        self.access_list = self.conf['access']['access_list']

        # -- REQUEST / RESPONSE
        self.path = self.conf['path_conf']['path']
        self.path_parameters = self.conf['path_conf']['parameters']

        # -- request query
        request_query_schema = self.conf['schemas'].get(
            'input_query') or {'schema': None, 'uri': None}

        self.request_query = Interface(
            self.name,
            Interface.TYPES.REQUEST_QUERY,
            request_query_schema['schema'],
            request_query_schema['uri'])

        # -- request body
        request_body_schema = self.conf['schemas'].get(
            'input_body') or {'schema': None, 'uri': None}

        self.request_body = Interface(
            self.name,
            Interface.TYPES.REQUEST_BODY,
            request_body_schema['schema'],
            request_body_schema['uri'])

        # -- response
        response_schema = self.conf['schemas']['output']
        self.response = Interface(
            self.name,
            Interface.TYPES.RESPONSE,
            response_schema['schema'],
            response_schema['uri'],
            bulk_read_field=self.get_bulk_read_field())
        self.signature = Signature(
            self.method,
            self.path,
            self.path_parameters,
            self.authorization_required,
            self.request_query,
            self.request_body,
            bulk_read_field=self.get_bulk_read_field())

        # -- EXAMPLES
        self.examples = self.conf['examples']

    @property
    def header(self):
        return normalize_indentation('''
        /**
         * {self.title}
         */
        ''', 0).format(self=self)

    def get_bulk_read_field(self):

        if not self.name.startswith('BULK_READ'):
            return None

        try:
            schema = self.conf['schemas']['output']['schema']
            fields = list(schema['properties'].keys())

        except KeyError:
            return None

        if len(fields) == 1:
            field = fields[0]
            if field and schema['properties'][field]['type'] == 'array':
                return field

        return None

    def render(self):

        if self.get_bulk_read_field() and self.method != 'get':
            return normalize_indentation('''
            {self.header}
            public {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                return this.client
                    .{self.method}<X.{self.response.name}>({self.signature.call_args})
                    .pipe(map(x => x.{bulk_read_field}));
            }}
            ''', 0).format(  # noqa
                self=self,
                bulk_read_field=self.get_bulk_read_field(),
            )

        elif self.get_bulk_read_field() or self.method == 'get':
            return normalize_indentation('''
            {self.header}
            public {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                return this.client.get<X.{self.response.name}>({self.signature.call_args});
            }}
            ''', 0).format(self=self)  # noqa

        else:
            return normalize_indentation('''
            {self.header}
            public {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                return this.client
                    .{self.method}<X.{self.response.name}>({self.signature.call_args})
                    .pipe(filter(x => !_.isEmpty(x)));
            }}
            ''', 0).format(self=self)  # noqa

    def render_facade(self):

        if self.get_bulk_read_field():
            return normalize_indentation('''
                {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                    return this.{self.domain_id}Domain.{self.camel_name}({self.signature.call_args_without_path});
                }}
            ''', 0).format(self=self)  # noqa

        elif self.method == 'get':
            return normalize_indentation('''
                {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                    return this.{self.domain_id}Domain.{self.camel_name}({self.signature.call_args_without_path});
                }}
            ''', 0).format(self=self)  # noqa

        else:
            return normalize_indentation('''
                {self.camel_name}({self.signature.input}): Observable<X.{self.response.name}> {{
                    return this.{self.domain_id}Domain.{self.camel_name}({self.signature.call_args_without_path});
                }}
            ''', 0).format(self=self)  # noqa

    def render_examples(self):

        examples_blocks = []
        for example_name in sorted(self.examples.keys()):
            example_response = self.examples[example_name]['response']

            examples_blocks.append(normalize_indentation('''
                "{example_name}": {example_response}
            ''', 0).format(
                example_name=example_name,
                example_response=json.dumps(
                    example_response, sort_keys=True, indent=4)))

        return normalize_indentation('''
            /**
             * Examples for {self.name}
             */
            export const {command_name}Examples = {{
            {examples}
            }}
        ''', 0).format(
            self=self,
            command_name=to_camelcase(self.name),
            examples=normalize_indentation(',\n\n'.join(examples_blocks), 4))

    def render_access(self, access_enum):

        if self.access_list:

            access_list = '[{}]'.format(
                ','.join([f'{access_enum}.{a}' for a in self.access_list]))

            return f'{self.name}: {access_list}'

        return f'{self.name}: null'




from unittest.mock import Mock

from django.test import TestCase
from django.db import models
import pytest

from lily.base import serializers


#
# Test Serializers
#
class Customer(models.Model):

    name = models.CharField(max_length=100)

    age = models.IntegerField(null=True, blank=True)

    is_ready = models.BooleanField()

    class Meta:
        app_label = 'base'


class CustomerSerializer(serializers.Serializer):

    _type = 'customer'

    name = serializers.CharField(max_length=123, required=False)

    age = serializers.IntegerField(min_value=18)

    def get_access(self, instance):

        return [
            (Mock(command_conf={'name': 'MARK_IT'}), True),
            (Mock(command_conf={'name': 'REMOVE_IT'}), False),
            ('ALLOW_IT', False),
        ]


class CustomerModelSerializer(serializers.ModelSerializer):

    _type = 'customer'

    def get_access(self, instance):

        return [
            (Mock(command_conf={'name': 'MARK_IT'}), True),
            (Mock(command_conf={'name': 'REMOVE_IT'}), False),
            ('ALLOW_IT', False),
        ]

    class Meta:
        model = Customer
        fields = ('name', 'age')


class SerializerTestCase(TestCase):

    serializer = CustomerSerializer

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker):
        self.mocker = mocker

    def test_to_internal_value(self):

        class MetaCustomer(serializers.Serializer):
            _type = 'customer'
            at__name = serializers.CharField()
            at__age = serializers.IntegerField()
            is_ready = serializers.BooleanField()

        s = MetaCustomer(data={
            '@name': 'George',
            '@age': 13,
            'is_ready': False,
        })

        assert s.is_valid() is True
        assert s.data == {
            '@type': 'customer',
            '@name': 'George',
            '@age': 13,
            'is_ready': False,
        }

    def test_to_internal_value_with_instance(self):

        class WhatSerializer(serializers.Serializer):
            what = serializers.IntegerField()

        class MetaCustomer(serializers.Serializer):
            _type = 'customer'
            name = serializers.CharField()
            age = serializers.IntegerField()
            is_ready = serializers.BooleanField()
            what = WhatSerializer()

        data = {
            'name': 'George',
            'age': 13,
            'is_ready': False,
            'what': ['what'],
        }
        s = MetaCustomer(data=data)

        assert dict(s.to_internal_value(data)) == {
            'name': 'George',
            'age': 13,
            'is_ready': False,
            'what': ['what'],
        }

    def test_to_representation__without_access(self):

        p = Customer(name='John', age=81)

        self.mocker.patch.object(
            self.serializer, 'get_access').return_value = []

        assert self.serializer(
            context={'request': Mock(), 'command_name': 'ACT_NOW'}
        ).to_representation(p) == {
            '@type': 'customer',
            'age': 81,
            'name': 'John',
        }

    def test_to_representation__with_access(self):

        p = Customer(name='John', age=81)

        assert self.serializer(
            context={'request': Mock(), 'command_name': 'ACT_NOW'}
        ).to_representation(p) == {
            '@type': 'customer',
            '@access': {
                'MARK_IT': True,
                'REMOVE_IT': False,
                'ALLOW_IT': False,
            },
            'age': 81,
            'name': 'John',
        }


class ModelSerializerTestCase(SerializerTestCase):

    serializer = CustomerModelSerializer


class EmptySerializerTestCase(TestCase):

    def test__passes_nothing(self):

        s = serializers.EmptySerializer({'hi': 'there'})

        assert s.data == {'@type': 'empty'}


class AbstractSerializerTestCase(TestCase):

    def test_type_not_needed(self):

        class AbstractCustomer(serializers.AbstractSerializer):

            name = serializers.CharField()

        assert AbstractCustomer({'name': 'Jack'}).data == {'name': 'Jack'}


class ObjectSerializerTestCase(TestCase):

    def test__passes_everything(self):

        s = serializers.ObjectSerializer({'hi': 'there'})

        assert s.data == {
            '@type': 'object',
            'hi': 'there',
        }


class CommandSerializerTestCase(TestCase):

    def test_serializes_command_correctly(self):

        s = serializers.CommandSerializer(data={
            'name': 'DO_IT',
            'method': 'post',
            'uri': 'http://do.it/now',
            'body': {'wat': 'yes'},
        })

        assert s.is_valid() is True
        assert s.data == {
            '@type': 'command',
            'name': 'DO_IT',
            'method': 'post',
            'uri': 'http://do.it/now',
            'body': {'wat': 'yes'},
        }

    def test__fails_on_invalid_input(self):
        s = serializers.CommandSerializer(data={
            'name': 'hi there',
            'uri': 'not url',
        })

        assert s.is_valid() is False
        assert s.errors == {
            'method': ['This field is required.'],
            'uri': ['Enter a valid URL.'],
        }




import os
from unittest import TestCase
from unittest.mock import call
import textwrap

from click.testing import CliRunner
import pytest

from lily.cli.cli import cli
from lily.cli.copier import Copier
from tests import remove_white_chars


class CliTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

    #
    # INIT
    #
    def test_init(self):

        copy = self.mocker.patch.object(Copier, 'copy')

        result = self.runner.invoke(cli, ['init', 'src_dir'])
        assert result.exit_code == 0
        assert result.output.strip() == textwrap.dedent('''
            [INFO]

            Please insert the following line at the top of your Makefile:

            include .lily/lily.makefile
        ''').strip()
        assert copy.call_args_list == [call('src_dir')]

    def test_init__config_does_not_exist(self):

        current_cwd = os.getcwd()
        os.chdir(str(self.tmpdir))

        copy = self.mocker.patch.object(Copier, 'copy')

        result = self.runner.invoke(cli, ['init', 'src_dir'])
        assert result.exit_code == 1
        assert remove_white_chars(result.output) == remove_white_chars('''
            Error:
            Please install `lily_assistant` and run
            `lily_assistant init <src_dir>` before running
            `lily init <src_dir>`
        ''')
        assert copy.call_count == 0
        os.chdir(current_cwd)




from unittest import TestCase
from time import time, sleep

from lily.async import AsyncTask, ParallelExecutor


class ParallelExecutorTestCase(TestCase):

    def test_execute__single_task(self):

        def success(x):
            sleep(0.5)
            return x ** 2

        t = AsyncTask(success, [4])

        ParallelExecutor(tasks=[t]).execute()

        assert t.result == 16
        assert t.successful is True

    def test_execute__many_tasks__all_successful(self):

        def success(x):
            sleep(0.5)
            return x ** 2

        t_0 = AsyncTask(success, [4])
        t_1 = AsyncTask(success, [6])
        t_2 = AsyncTask(success, [8])
        t_3 = AsyncTask(success, [9])

        start = time()
        ParallelExecutor(tasks=[t_0, t_1, t_2, t_3]).execute()
        end = time() - start

        assert end < 1
        assert t_0.result == 16
        assert t_0.successful is True
        assert t_1.result == 36
        assert t_1.successful is True
        assert t_2.result == 64
        assert t_2.successful is True
        assert t_3.result == 81
        assert t_3.successful is True

    def test_execute__many_tasks__one_failed(self):

        def success(x):
            sleep(0.5)
            return x ** 2

        def fail(x):
            raise Exception(x)

        t_0 = AsyncTask(success, [4])
        t_1 = AsyncTask(success, [6])
        t_2 = AsyncTask(fail, [8])

        start = time()
        ParallelExecutor(tasks=[t_0, t_1, t_2]).execute()
        end = time() - start

        assert end < 1
        assert t_0.result == 16
        assert t_0.successful is True
        assert t_1.result == 36
        assert t_1.successful is True
        assert isinstance(t_2.result, Exception)
        assert t_2.successful is False

    def test_execute__many_tasks__all_failed(self):

        def fail(x):
            raise Exception(x)

        t_0 = AsyncTask(fail, [4])
        t_1 = AsyncTask(fail, [6])
        t_2 = AsyncTask(fail, [8])

        start = time()
        ParallelExecutor(tasks=[t_0, t_1, t_2]).execute()
        end = time() - start

        assert end < 1
        assert isinstance(t_0.result, Exception)
        assert t_0.successful is False
        assert isinstance(t_1.result, Exception)
        assert t_1.successful is False
        assert isinstance(t_2.result, Exception)
        assert t_2.successful is False




import logging
import re

from . import serializers
from .events import EventFactory


logger = logging.getLogger()


class Domain(EventFactory):

    def __init__(self, id, name):
        if re.search(r'\s+', id):
            raise self.BrokenRequest(
                'BROKEN_ARGS_DETECTED',
                data={
                    'errors': {
                        'id': [
                            'should not contain white characters.',
                        ]
                    }
                })
        self.id = id.lower()
        self.name = name

    def __eq__(self, other):
        return (
            self.id == other.id and
            self.name == other.name)


class DomainSerializer(serializers.Serializer):

    _type = 'domain'

    id = serializers.CharField()

    name = serializers.CharField()


class Meta:

    def __init__(self, title, domain, description=None):
        self.title = title
        self.description = self.transform_description(description)
        self.domain = domain

    def __eq__(self, other):
        return (
            self.title == other.title and
            self.description == other.description and
            self.domain == other.domain)

    def transform_description(self, description):

        if description:
            description = description.strip()
            description = re.sub(r'[\n ]+', ' ', description)

        return description


class MetaSerializer(serializers.Serializer):

    _type = 'meta'

    title = serializers.CharField()

    description = serializers.CharField()

    domain = DomainSerializer()




from django.test import TestCase
import pytest

from lily.docs.renderers.angular.interface import Interface, Enum
from lily.base.utils import normalize_indentation
from lily import EventFactory


class EnumTestCase(TestCase):

    #
    # NAME
    #
    def test_name(self):
        assert Enum('age', []).name == 'Age'
        assert Enum('a_ge', []).name == 'AGe'

    def test_name__empty_name(self):

        with pytest.raises(EventFactory.BrokenRequest) as e:
            Enum('', ['what'])

        assert e.value.data == {
            '@event': 'ENUMS_WITHOUT_NAME_DETECTED',
            '@type': 'error',
            'name': '',
            'values': ['what'],
        }

    #
    # RENDER
    #
    def test_render(self):

        enum = Enum('age', ['AA', 'BB'])

        assert enum.render() == normalize_indentation('''
            export enum Age {
                AA = 'AA',
                BB = 'BB',
            }
        ''', 0)

    def test_render__numerical(self):

        enum0 = Enum('position', [0, 1])
        enum1 = Enum('position', ['0', '1', '2'])

        assert enum0.render() == normalize_indentation('''
            export enum Position {
                VALUE_0 = 0,
                VALUE_1 = 1,
            }
        ''', 0)
        assert enum1.render() == normalize_indentation('''
            export enum Position {
                VALUE_0 = '0',
                VALUE_1 = '1',
                VALUE_2 = '2',
            }
        ''', 0)

    def test_render__starts_from_number(self):

        enum = Enum('quality', ['240p', '720p', '1080p'])

        assert enum.render() == normalize_indentation('''
            export enum Quality {
                VALUE_1080P = '1080p',
                VALUE_240P = '240p',
                VALUE_720P = '720p',
            }
        ''', 0)

    def test_render__contains_extra_characters(self):

        enum = Enum(
            'content_type',
            ['image/svg+xml', 'video/png', 'audio/ogg-what'])

        assert enum.render() == normalize_indentation('''
            export enum ContentType {
                AUDIO_OGG_WHAT = 'audio/ogg-what',
                IMAGE_SVG_XML = 'image/svg+xml',
                VIDEO_PNG = 'video/png',
            }
        ''', 0)

    def test_render__removes_duplicates(self):

        enum = Enum('class', ['AA', 'BB', 'AA'])

        assert enum.render() == normalize_indentation('''
            export enum Class {
                AA = 'AA',
                BB = 'BB',
            }
        ''', 0)

    def test_render__sorts_values(self):

        enum = Enum('category', ['XX', 'AA', 'BB'])

        assert enum.render() == normalize_indentation('''
            export enum Category {
                AA = 'AA',
                BB = 'BB',
                XX = 'XX',
            }
        ''', 0)

    #
    # EQUAL_FIELD_AND_VALUES
    #
    def test__eq__(self):

        # -- identical
        assert (
            Enum('category', ['XX', 'AA', 'BB']) ==
            Enum('category', ['XX', 'AA', 'BB']))

        # -- the same, but with some duplicates and different order
        assert (
            Enum('category', ['XX', 'AA', 'BB']) ==
            Enum('category', ['XX', 'BB', 'AA', 'BB']))

        # -- different name
        assert (
            Enum('categories', ['XX', 'AA']) !=
            Enum('category', ['XX', 'AA']))

        # -- different values
        assert (
            Enum('category', ['XX', 'AA']) !=
            Enum('category', ['XX', 'YY']))


class InterfaceTestCase(TestCase):

    def test_name(self):

        assert Interface(
            'READ_CARDS', Interface.TYPES.RESPONSE, {'hi': 'there'}, ''
        ).name == 'ReadCardsResponse'

        assert Interface(
            'READ_CARDS', Interface.TYPES.REQUEST_QUERY, {'hi': 'there'}, ''
        ).name == 'ReadCardsQuery'

        assert Interface(
            'READ_CARDS', Interface.TYPES.REQUEST_BODY, {'hi': 'there'}, ''
        ).name == 'ReadCardsBody'

    def test_is_empty(self):

        assert Interface('X', 'X', None, '').is_empty() is True
        assert Interface('X', 'X', {}, '').is_empty() is True
        assert Interface('X', 'X', {'a': 'b'}, '').is_empty() is False

    def test_append_enum__no_duplicates(self):

        interface = Interface('READ_CARDS', Interface.TYPES.RESPONSE, {}, '')

        e0 = interface.append_enum('age', [19, 91])

        assert len(interface.enums) == 1
        assert e0 == Enum('age', [19, 91], None)

    def test_append_enum__with_duplicates(self):

        interface = Interface('READ_CARDS', Interface.TYPES.RESPONSE, {}, '')

        e0 = interface.append_enum('age', [19, 91])
        e1 = interface.append_enum('age', [19, 91])

        assert len(interface.enums) == 1
        assert e0 == e1

    def test_append_enum__same_values_different_names(self):

        interface = Interface('READ_CARDS', Interface.TYPES.RESPONSE, {}, '')

        e0 = interface.append_enum('age', [19, 91])  # noqa
        e1 = interface.append_enum('years', [19, 91])  # noqa

        assert len(interface.enums) == 2

    def test_append_enum__same_name_different_values(self):

        interface = Interface('X', Interface.TYPES.RESPONSE, {}, '')

        e0 = interface.append_enum('age', [19, 91])  # noqa
        e1 = interface.append_enum('age', [19])

        assert len(interface.enums) == 2
        assert e1 == Enum('age', [19], 1)


@pytest.mark.parametrize(
    'schema, bulk_read_field, expected_rendered, expected_enums',
    [
        # -- case 0 - None schema
        (
            None,
            None,
            '',
            [],
        ),

        # -- case 1 - {} schema
        (
            {},
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {}
            ''', 0),
            [],
        ),

        # -- case 2 - empty schema
        (
            {'type': 'object', 'properties': {}},
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
            }
            ''', 0),
            [],
        ),

        # -- case 3 - simple schema
        (
            {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'integer',
                        'minimum': 18,
                    },
                    'name': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                age: number;
                name?: string;
            }
            ''', 0),
            [],
        ),

        # -- case 4 - simple array
        (
            {
                'type': 'object',
                'required': ['age'],
                'entity_type': 'human',
                'properties': {
                    'age': {
                        'type': 'array',
                        'items': {
                            'type': 'integer'
                        },
                    },
                    'surname': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                '@type'?: 'human';
                '@access'?: {
                    [name: string]: boolean;
                };
                age: number[];
                surname?: string;
            }
            ''', 0),
            [],
        ),

        # -- case 5 - enum
        (
            {
                'type': 'object',
                'required': ['occupation'],
                'properties': {
                    'occupation': {
                        'type': 'string',
                        'enum_name': 'occupation_type',
                        'enum': ['AA', 'BB'],
                    },
                    'surname': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                occupation: OccupationType;
                surname?: string;
            }
            ''', 0),
            [Enum('OccupationType', ['AA', 'BB'])],
        ),

        # -- case 6 - enum with const value
        (
            {
                'type': 'object',
                'properties': {
                    'employees': {
                        'type': 'array',
                        'items': {
                            'oneOf': [
                                {
                                    'type': 'object',
                                    'properties': {
                                        'name': {
                                            'type': 'string',
                                        },
                                        'occupation': {
                                            'type': 'string',
                                            'enum_name': 'occupation_type',
                                            'enum': ['AA', 'BB'],
                                            'const_value': 'AA',
                                        }
                                    },
                                    'required': ['occupation'],
                                },
                                {
                                    'type': 'object',
                                    'properties': {
                                        'occupation': {
                                            'type': 'string',
                                            'enum_name': 'occupation_type',
                                            'enum': ['AA', 'BB'],
                                            'const_value': 'BB',
                                        }
                                    },
                                    'required': ['occupation'],
                                },
                            ]
                        },
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                employees?: ({
                    name?: string;
                    occupation: OccupationType.AA;
                } | {
                    occupation: OccupationType.BB;
                })[];
            }

            ''', 0),
            [Enum('OccupationType', ['AA', 'BB'])],
        ),

        # -- case 7 - enums
        (
            {
                'type': 'object',
                'required': ['occupation'],
                'properties': {
                    'occupation': {
                        'type': 'string',
                        'enum': ['AA', 'BB'],
                        'enum_name': 'occupation',
                    },
                    'age': {
                        'type': 'string',
                        'enum': ['12', '21', '33'],
                        'enum_name': 'AgeChoice',
                    },
                    'surname': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                age?: AgeChoice;
                occupation: Occupation;
                surname?: string;
            }
            ''', 0),
            [
                Enum('AgeChoice', ['12', '21', '33']),
                Enum('Occupation', ['AA', 'BB']),
            ],
        ),

        # -- case 8 - enums array
        (
            {
                'type': 'object',
                'required': ['occupation'],
                'properties': {
                    'occupation': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'enum_name': 'occupation',
                            'enum': ['AA', 'BB'],
                        }
                    },
                    'surname': {
                        'maxLength': 123,
                        'type': 'string',
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                occupation: Occupation[];
                surname?: string;
            }
            ''', 0),
            [Enum('Occupation', ['AA', 'BB'])],
        ),

        # -- case 9 - 1 level deep nested
        (
            {
                'type': 'object',
                'required': ['age'],
                'entity_type': 'employee',
                'properties': {
                    'age': {
                        'type': 'integer',
                    },
                    'person': {
                        'type': 'object',
                        'required': ['name'],
                        'entity_type': 'human',
                        'properties': {
                            'name': {
                                'type': 'string',
                            },
                            'category': {
                                'type': 'number'
                            }
                        },
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                '@type'?: 'employee';
                '@access'?: {
                    [name: string]: boolean;
                };
                age: number;
                person?: {
                    '@type'?: 'human';
                    '@access'?: {
                        [name: string]: boolean;
                    };
                    category?: number;
                    name: string;
                };
            }
            ''', 0),
            [],
        ),

        # -- case 10 - 2 levels deep nested
        (
            {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'integer',
                    },
                    'person': {
                        'type': 'object',
                        'required': ['name'],
                        'properties': {
                            'name': {
                                'type': 'string',
                            },
                            'category': {
                                'type': 'object',
                                'required': ['id'],
                                'properties': {
                                    'id': {
                                        'type': 'integer',
                                    },
                                    'name': {
                                        'type': 'string',
                                    }
                                }
                            }
                        },
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                age: number;
                person?: {
                    category?: {
                        id: number;
                        name?: string;
                    };
                    name: string;
                };
            }
            ''', 0),
            [],
        ),

        # -- case 11 - nested array
        (
            {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'integer',
                    },
                    'people': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'required': ['name'],
                            'properties': {
                                'name': {
                                    'type': 'string',
                                },
                                'category': {
                                    'type': 'number'
                                }
                            },
                        },
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                age: number;
                people?: {
                    category?: number;
                    name: string;
                }[];
            }
            ''', 0),
            [],
        ),

        # -- case 12 - nested with enums
        (
            {
                'type': 'object',
                'required': ['age'],
                'properties': {
                    'age': {
                        'type': 'string',
                        'enum': ['A', 'B'],
                        'enum_name': 'MyAge',
                    },
                    'people': {
                        'type': 'object',
                        'required': ['category'],
                        'properties': {
                            'category': {
                                'type': 'string',
                                'enum_name': 'cat',
                                'enum': ['CAT0', 'CAT1', 'CAT2'],
                            }
                        },
                    },
                },
            },
            None,
            normalize_indentation('''
            /**
             * http://here
             */

            export interface ReadCardsResponse {
                age: MyAge;
                people?: {
                    category: Cat;
                };
            }
            ''', 0),
            [
                Enum('Cat', ['CAT0', 'CAT1', 'CAT2']),
                Enum('MyAge', ['A', 'B']),
            ],
        ),

        # -- case 13 - mixed up primitive types
        (
            {
                'required': ['attempt_stats', 'count'],
                'type': 'object',
                'properties': {
                    'count': {'type': 'integer'},
                    'attempt_stats': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'required': ['created_timestamp'],
                            'properties': {
                                'created_timestamp': {
                                    'type': 'number'
                                },
                                'card_id': {
                                    'type': 'any'
                                },
                                'states_path': {
                                    'type': 'object'
                                },
                                'id': {
                                    'type': 'integer'
                                },
                                'successful': {
                                    'type': 'boolean'
                                }
                            },
                        }
                    }
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    attempt_stats: {
                        card_id?: any;
                        created_timestamp: number;
                        id?: number;
                        states_path?: any;
                        successful?: boolean;
                    }[];
                    count: number;
                }

            ''', 0),
            [],
        ),

        # -- case 14 - const types
        (
            {
                'required': ['type', 'age'],
                'type': 'object',
                'properties': {
                    'type': {
                        'type': 'string',
                        'pattern': '^MAN$',
                        'const': 'MAN',
                    },
                    'age': {
                        'type': 'number',
                    }
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    age: number;
                    type: "MAN";
                }

            ''', 0),
            [],
        ),

        # -- case 15 - const and oneOf
        (
            {
                'required': ['person'],
                'type': 'object',
                'properties': {
                    'person': {
                        'oneOf': [
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'type': 'string',
                                        'pattern': '^DOG$',
                                        'const': 'DOG',
                                    },
                                    'age': {
                                        'type': 'number',
                                    }
                                }
                            },
                            {
                                'type': 'object',
                                'properties': {
                                    'type': {
                                        'type': 'string',
                                        'pattern': '^CAT$',
                                        'const': 'CAT',
                                    },
                                    'name': {
                                        'type': 'string',
                                    }
                                }
                            },
                        ],
                    }
                }
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    person: {
                        age?: number;
                        type?: "DOG";
                    } | {
                        name?: string;
                        type?: "CAT";
                    };
                }
            ''', 0),
            [],
        ),

        # -- case 16 - bulk read response
        (
            {
                'type': 'object',
                'required': ['people'],
                'properties': {
                    'people': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'required': ['name', 'age'],
                            'properties': {
                                'name': {
                                    'type': 'string',
                                },
                                'age': {
                                    'type': 'integer',
                                },
                            },
                        },
                    },
                },
            },
            'cards',
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponseEntity {
                    age: number;
                    name: string;
                }

                export interface ReadCardsResponse {
                    cards: ReadCardsResponseEntity[];
                }
            ''', 0),
            [],
        ),

        # -- case 17 - bulk read simple response - integer
        (
            {
                'type': 'object',
                'required': [
                    'card_ids'
                ],
                'properties': {
                    'card_ids': {
                        'type': 'array',
                        'items': {
                            'type': 'integer'
                        }
                    }
                }
            },
            'card_ids',
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponseEntity extends Number {}

                export interface ReadCardsResponse {
                    card_ids: ReadCardsResponseEntity[];
                }
            ''', 0),
            [],
        ),

        # -- case 18 - bulk read simple response - boolean
        (
            {
                'type': 'object',
                'required': [
                    'card_ids'
                ],
                'properties': {
                    'card_ids': {
                        'type': 'array',
                        'items': {
                            'type': 'boolean'
                        }
                    }
                }
            },
            'card_ids',
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponseEntity extends Boolean {}

                export interface ReadCardsResponse {
                    card_ids: ReadCardsResponseEntity[];
                }
            ''', 0),
            [],
        ),

        # -- case 19 - bulk read simple response - string
        (
            {
                'type': 'object',
                'required': [
                    'card_ids'
                ],
                'properties': {
                    'card_ids': {
                        'type': 'array',
                        'items': {
                            'type': 'string'
                        }
                    }
                }
            },
            'card_ids',
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponseEntity extends String {}

                export interface ReadCardsResponse {
                    card_ids: ReadCardsResponseEntity[];
                }
            ''', 0),
            [],
        ),

        # -- case 20 - bulk read response
        (
            {
                'type': 'object',
                'required': ['people'],
                'properties': {
                    'people': {
                        'type': 'object',
                    },
                },
            },
            'cards',
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponseEntity extends Object {};

                export interface ReadCardsResponse {
                    cards: ReadCardsResponseEntity[];
                }
            ''', 0),
            [],
        ),

        # -- case 21 - one of - null or string
        (
            {
                'type': 'object',
                'properties': {
                    'name': {
                        'oneOf': [
                            {'type': 'null'},
                            {'type': 'string'},
                        ],
                    },
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    name?: null | string;
                }
            ''', 0),
            [],
        ),

        # -- case 23 - one of - null or object
        (
            {
                'type': 'object',
                'required': ['background'],
                'properties': {
                    'background': {
                        'oneOf': [
                            {'type': 'null'},
                            {
                                'type': 'object',
                                'properties': {
                                    'audio_uri': {
                                        'type': 'string',
                                    },
                                    'audio_stop': {
                                        'type': 'number',
                                    },
                                },
                            },
                        ],
                    },
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    background: null | {
                        audio_stop?: number;
                        audio_uri?: string;
                    };
                }
            ''', 0),
            [],
        ),

        # -- case 24 - one of - null or object or string or other object
        (
            {
                'type': 'object',
                'required': ['background'],
                'properties': {
                    'background': {
                        'oneOf': [
                            {'type': 'null'},
                            {
                                'type': 'object',
                                'properties': {
                                    'audio_uri': {
                                        'type': 'string',
                                    },
                                    'audio_stop': {
                                        'type': 'number',
                                    },
                                },
                            },
                            {'type': 'string'},
                            {
                                'type': 'object',
                                'properties': {
                                    'text': {
                                        'type': 'string',
                                    },
                                    'age': {
                                        'type': 'number',
                                    },
                                },
                            },
                        ],
                    },
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    background: null | {
                        audio_stop?: number;
                        audio_uri?: string;
                    } | string | {
                        age?: number;
                        text?: string;
                    };
                }
            ''', 0),
            [],
        ),

        # -- case 25 - one of - null or complex object
        (
            {
                'type': 'object',
                'required': ['background'],
                'properties': {
                    'background': {
                        'oneOf': [
                            {'type': 'null'},
                            {
                                'type': 'object',
                                'properties': {
                                    'audio_uri': {
                                        'oneOf': [
                                            {'type': 'null'},
                                            {
                                                'type': 'string',
                                                'format': 'url',
                                            },
                                        ],
                                    },
                                    'audio_text': {
                                        'oneOf': [
                                            {'type': 'null'},
                                            {'type': 'string'},
                                        ],
                                    },
                                    'audio_language': {
                                        'oneOf': [
                                            {'type': 'null'},
                                            {
                                                'type': 'string',
                                                'enum_name': 'language',
                                                'enum': [
                                                    'fr',
                                                    'nb',
                                                    'is',
                                                    'en',
                                                ],
                                            },
                                        ],
                                    },
                                    'audio_stop': {
                                        'oneOf': [
                                            {'type': 'null'},
                                            {'type': 'number'},
                                        ],
                                    },
                                },
                            },
                        ],
                    },
                },
            },
            None,
            normalize_indentation('''
                /**
                 * http://here
                 */

                export interface ReadCardsResponse {
                    background: null | {
                        audio_language?: null | Language;
                        audio_stop?: null | number;
                        audio_text?: null | string;
                        audio_uri?: null | string;
                    };
                }

            ''', 0),
            [Enum('Language', ['fr', 'is', 'nb', 'en'])],
        ),
    ], ids=list([str(i) for i in range(25)]))
def test_render(schema, bulk_read_field, expected_rendered, expected_enums):

    rendered, enums = Interface(
        'READ_CARDS',
        Interface.TYPES.RESPONSE,
        schema,
        'http://here',
        bulk_read_field=bulk_read_field).render()

    assert rendered == expected_rendered
    assert (
        sorted(enums, key=lambda x: x.name) ==
        sorted(expected_enums, key=lambda x: x.name))




import re

from .utils import to_camelcase
from lily.base.utils import normalize_indentation
from lily import EventFactory


class Interface:

    class TYPES:  # noqa
        RESPONSE = 'RESPONSE'

        REQUEST_QUERY = 'REQUEST_QUERY'

        REQUEST_BODY = 'REQUEST_BODY'

    def __init__(self, command_name, type_, schema, uri, bulk_read_field=None):
        self.command_name = command_name
        self.type = type_
        self.schema = schema
        self.uri = uri
        self.enums = []
        self.bulk_read_field = bulk_read_field

    def is_empty(self):
        return self.schema is None or self.schema == {}

    def append_enum(self, name, values):

        new_enum = Enum(name, values)
        if new_enum not in self.enums:
            self.enums.append(new_enum)

        return new_enum

    @property
    def name(self):
        if self.type == self.TYPES.RESPONSE:
            suffix = 'Response'

        elif self.type == self.TYPES.REQUEST_QUERY:
            suffix = 'Query'

        else:
            suffix = 'Body'

        return '{name}{suffix}'.format(
            name=to_camelcase(self.command_name),
            suffix=suffix)

    def render(self):

        if self.schema is None:
            return '', []

        def optional(name, schema):
            return name not in schema.get('required', []) and '?' or ''

        def to_type(name, schema, indent):
            _type = schema.get('type')
            if _type in ['integer', 'number']:
                return 'number'

            elif _type == 'string' and schema.get('enum'):
                enum = self.append_enum(schema['enum_name'], schema['enum'])

                if schema.get('const_value'):
                    v = schema['const_value']
                    return f'{enum.name}.{enum.get_key_for_value(v)}'

                return enum.name

            elif _type == 'null':
                return 'null'

            elif _type == 'boolean':
                return 'boolean'

            elif _type == 'string':
                if 'const' in schema:
                    const = schema['const']
                    return f'"{const}"'

                return 'string'

            elif _type == 'any':
                return 'any'

            elif _type == 'object':
                if 'properties' in schema:
                    return to_interface(
                        schema, indent=indent + 4, base_indent=indent)
                else:
                    return 'any'

            return 'any'

        def to_interface(schema, indent=4, base_indent=0):

            lines = []

            if schema['type'] == 'object':
                lines.append('{')

                entity_type = schema.get('entity_type')
                if entity_type:
                    lines.append(
                        normalize_indentation(
                            """
                                '@type'?: '{entity_type}';
                                '@access'?: {{
                                    [name: string]: boolean;
                                }};
                            """.format(entity_type=entity_type),
                            indent))

                names = sorted(schema['properties'].keys())

                for name in names:
                    sub_schema = schema['properties'][name]

                    if sub_schema.get('oneOf'):
                        values = []
                        for alt_schema in sub_schema.get('oneOf'):
                            values.append(to_type(name, alt_schema, indent))

                        value = ' | '.join(values)

                    elif (sub_schema.get('type') == 'array' and
                            sub_schema['items'].get('oneOf')):

                        values = []
                        for alt_schema in sub_schema['items']['oneOf']:
                            values.append(to_type(name, alt_schema, indent))

                        value = ' | '.join(values)
                        value = f'({value})[]'

                    elif sub_schema.get('type') == 'array':
                        value = to_type(name, sub_schema['items'], indent)
                        value += '[]'

                    else:
                        value = to_type(name, sub_schema, indent)

                    lines.append(
                        '{indent}{name}{optional}: {value};'.format(
                            indent=' ' * indent,
                            name=name,
                            optional=optional(name, schema),
                            value=value)
                    )

                lines.append(' ' * base_indent + '}')

            elif schema['type'] in ['integer', 'number']:
                lines.append('extends Number {}')

            elif schema['type'] in ['string', 'boolean']:
                lines.append(
                    'extends {} {{}}'.format(schema['type'].capitalize()))

            return '\n'.join(lines)

        blocks = [normalize_indentation('''
            /**
             * {self.uri}
             */
        ''', 0).format(self=self)]

        # -- treat differently Bulk Read Response interfaces in order to
        # -- allow one extra mapping in the domain service
        if self.bulk_read_field:
            try:
                field = list(self.schema['properties'].keys())[0]
                entity_schema = self.schema['properties'][field]['items']

                interface = normalize_indentation('''
                    export interface {self.name}Entity {interface_content}

                    export interface {self.name} {{
                        {self.bulk_read_field}: {self.name}Entity[];
                    }}

                    ''', 0).format(
                    self=self,
                    interface_content=to_interface(entity_schema))

            # -- FIXME: this is a hack allowing one to provide very
            # -- forgiving interface when output schema is missing, when
            # -- for example one uses output from one service to provide
            # -- output for itself. This behaviour will be solved in the future
            # -- by usage of service client with exportable serializes
            except KeyError:

                interface = normalize_indentation('''
                    export interface {self.name}Entity extends Object {{}};

                    export interface {self.name} {{
                        {self.bulk_read_field}: {self.name}Entity[];
                    }}

                    ''', 0).format(self=self)

        # -- deal with empty schema which might occur for the case of
        # -- responses that return only @event field
        elif not self.schema:
            interface = 'export interface {self.name} {{}}'.format(self=self)

        else:
            interface = normalize_indentation(
                'export interface {self.name} {interface_content}'.format(
                    self=self,
                    interface_content=to_interface(self.schema)), 0)

        blocks.append(interface)

        return '\n\n'.join(blocks), self.enums


class Enum:

    def __init__(self, name, values, index=None):
        if not name:
            raise EventFactory.BrokenRequest(
                'ENUMS_WITHOUT_NAME_DETECTED',
                data={'name': name, 'values': values})

        self.name = to_camelcase(name)
        self.values = set(values)
        self.index = index

    def __eq__(self, other):
        return (
            self.name.lower() == other.name.lower() and
            self.values == other.values)

    def render(self):

        values = self.values
        value_pairs = []
        for v in sorted(values):

            if isinstance(v, int):
                value = v

            else:
                value = f"'{v}'"

            value_pairs.append((self.get_key_for_value(v), value))

        return 'export enum {name} {{\n{fields}\n}}'.format(
            name=self.name,
            fields='\n'.join([
                f"    {k} = {v}," for k, v in value_pairs]))

    def get_key_for_value(self, value):

        def transform_key(k):
            return re.sub(r'[^\w]+', '_', k).upper()

        if isinstance(value, int):
            key = f'VALUE_{value}'

        elif isinstance(value, str) and re.search(r'^\d+', value):
            key = f'VALUE_{value}'

        else:
            key = value

        return transform_key(key)




import os

from lily_assistant.config import Config
import djclick as click


@click.command()
def command():

    examples_path = os.path.join(Config.get_lily_path(), 'examples.json')

    try:
        os.remove(examples_path)

    except FileNotFoundError:
        pass

    click.echo("'examples.json' was removed")




from click.testing import CliRunner
from django.test import TestCase
import pytest

from lily_assistant.config import Config

from lily.docs.management.commands.render_markdown import command
from lily.docs.renderers.markdown.renderer import MarkdownRenderer


class RenderMarkdownTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

    def test_command(self):

        lily_dir = self.tmpdir.mkdir('.lily')
        self.mocker.patch.object(
            Config, 'get_lily_path').return_value = str(lily_dir)

        self.mocker.patch.object(
            MarkdownRenderer, 'render').return_value = '# API'

        result = self.runner.invoke(command, [])

        assert result.exit_code == 0
        assert result.output.strip() == (
            'Successfully rendered Markdown Specification for all '
            'registered Commands')
        assert lily_dir.join('API.md').read() == '# API'




import json
from collections import OrderedDict
from unittest.mock import Mock

from django.test import TestCase
import pytest

from lily.entrypoint.renderer import CommandsRenderer
from lily.entrypoint.base import BaseRenderer
from lily import Input, Output, Meta, Domain, Access, Source


def fn():
    pass


class CommandsRendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

        examples = self.tmpdir.join('examples.json')
        examples.write('{}')
        self.examples_filepath = str(examples)
        self.mocker.patch(
            'lily.entrypoint.renderer.get_examples_filepath'
        ).return_value = self.examples_filepath

    def test_render(self):
        renderer = self.mocker.patch(
            'lily.entrypoint.renderer.SchemaRenderer')
        serialize = Mock()
        renderer.return_value.render.return_value = Mock(
            serialize=serialize, enums=[])
        serialize.side_effect = [
            {'output': 'schema'},
            {'query': 'schema'},
            {'body': 'schema'},
        ]

        meta = Meta(
            title='hi',
            description='ho',
            domain=Domain(id='h', name='hh'))
        access = Access(access_list=['EVERYONE'], is_private=True)
        source = Source(fn)
        self.mocker.patch.object(BaseRenderer, 'render').return_value = {
            'READ_CARD': {
                'method': 'get',
                'path_conf': {
                    'path': '/hi',
                    'pattern': '/hi',
                    'parameters': [],
                },
                'meta': meta,
                'access': access,
                'input': Input(query_parser=Mock(), body_parser=Mock()),
                'output': Output(serializer=Mock()),
                'source': source,
            }
        }

        assert CommandsRenderer().render() == {
            '@enums': [],
            'READ_CARD': {
                'access': access,
                'meta': meta,
                'method': 'get',
                'path_conf': {
                    'parameters': [],
                    'path': '/hi',
                },
                'schemas': {
                    'input_body': {'body': 'schema'},
                    'input_query': {'query': 'schema'},
                    'output': {'output': 'schema'},
                },
                'source': source,
                'examples': {},
            }
        }

    def test_render__with_markdown_description(self):
        renderer = self.mocker.patch(
            'lily.entrypoint.renderer.SchemaRenderer')
        serialize = Mock()
        renderer.return_value.render.return_value = Mock(
            serialize=serialize, enums=[])
        serialize.side_effect = [
            {'output': 'schema'},
            {'query': 'schema'},
            {'body': 'schema'},
        ]

        meta = Meta(
            title='hi',
            description='./description.md',
            domain=Domain(id='h', name='hh'))
        access = Access(access_list=['EVERYONE'], is_private=True)
        source = Source(fn)
        self.mocker.patch.object(BaseRenderer, 'render').return_value = {
            'READ_CARD': {
                'method': 'get',
                'path_conf': {
                    'path': '/hi',
                    'pattern': '/hi',
                    'parameters': [],
                },
                'meta': meta,
                'access': access,
                'input': Input(query_parser=Mock(), body_parser=Mock()),
                'output': Output(serializer=Mock()),
                'source': source,
            }
        }

        result = CommandsRenderer().render()

        meta = Meta(
            title='hi',
            description='# this is test of markdown description',
            domain=Domain(id='h', name='hh'))
        assert result == {
            '@enums': [],
            'READ_CARD': {
                'access': access,
                'meta': meta,
                'method': 'get',
                'path_conf': {
                    'parameters': [],
                    'path': '/hi',
                },
                'schemas': {
                    'input_body': {'body': 'schema'},
                    'input_query': {'query': 'schema'},
                    'output': {'output': 'schema'},
                },
                'source': source,
                'examples': {},
            }
        }

    def test_render__with_enums(self):

        renderer = self.mocker.patch(
            'lily.entrypoint.renderer.SchemaRenderer')
        renderer.return_value.render.side_effect = [
            Mock(
                enums=[{'enum_name': 'A'}],
                serialize=Mock(return_value={'output': 'schema'})),
            Mock(
                enums=[
                    {'enum_name': 'C'},
                    {'enum_name': 'A'},
                    {'enum_name': 'D'},
                ],
                serialize=Mock(return_value={'query': 'schema'})),
            Mock(
                enums=[{'enum_name': 'D'}],
                serialize=Mock(return_value={'body': 'schema'})),
        ]

        meta = Meta(
            title='hi',
            description='ho',
            domain=Domain(id='h', name='hh'))
        access = Access(access_list=['EVERYONE'], is_private=True)
        source = Source(fn)
        self.mocker.patch.object(BaseRenderer, 'render').return_value = {
            'READ_CARD': {
                'method': 'get',
                'path_conf': {
                    'path': '/hi',
                    'pattern': '/hi',
                    'parameters': [],
                },
                'meta': meta,
                'access': access,
                'input': Input(query_parser=Mock(), body_parser=Mock()),
                'output': Output(serializer=Mock()),
                'source': source,
            }
        }

        assert CommandsRenderer().render() == {
            '@enums': [
                {'enum_name': 'A'},
                {'enum_name': 'C'},
                {'enum_name': 'D'},
            ],
            'READ_CARD': {
                'access': access,
                'meta': meta,
                'method': 'get',
                'path_conf': {
                    'parameters': [],
                    'path': '/hi',
                },
                'schemas': {
                    'input_body': {'body': 'schema'},
                    'input_query': {'query': 'schema'},
                    'output': {'output': 'schema'},
                },
                'source': source,
                'examples': {},
            }
        }

    def test_render__many_commands(self):
        renderer = self.mocker.patch(
            'lily.entrypoint.renderer.SchemaRenderer')
        serialize = Mock()
        renderer.return_value.render.return_value = Mock(
            serialize=serialize, enums=[])
        serialize.side_effect = [
            {'output': 'read.schema'},
            {'query': 'read.schema'},
            {'body': 'read.schema'},
            {'output': 'delete.schema'},
            {'query': 'delete.schema'},
            {'body': 'delete.schema'},
        ]

        meta = Meta(
            title='hi',
            description='ho',
            domain=Domain(id='h', name='hh'))
        access = Access(access_list=['EVERYONE'], is_private=True)
        source = Source(fn)
        self.mocker.patch.object(
            BaseRenderer,
            'render'
        ).return_value = OrderedDict([
            (
                'READ_CARD',
                {
                    'method': 'get',
                    'path_conf': {
                        'path': '/hi',
                        'pattern': '/hi',
                        'parameters': [],
                    },
                    'meta': meta,
                    'access': access,
                    'input': Input(query_parser=Mock(), body_parser=Mock()),
                    'output': Output(serializer=Mock()),
                    'source': source,
                },
            ),
            (
                'DELETE_TASK',
                {
                    'method': 'delete',
                    'path_conf': {
                        'path': '/hi/{id}',
                        'pattern': '/hi/(?P<id>\\d+)',
                        'parameters': [{'name': 'id', 'type': 'integer'}],
                    },
                    'meta': meta,
                    'access': access,
                    'input': Input(query_parser=Mock(), body_parser=Mock()),
                    'output': Output(serializer=Mock()),
                    'source': source,
                }
            )])

        assert CommandsRenderer().render() == {
            '@enums': [],
            'READ_CARD': {
                'access': access,
                'meta': meta,
                'method': 'get',
                'path_conf': {
                    'parameters': [],
                    'path': '/hi',
                },
                'schemas': {
                    'input_body': {'body': 'read.schema'},
                    'input_query': {'query': 'read.schema'},
                    'output': {'output': 'read.schema'},
                },
                'source': source,
                'examples': {},
            },
            'DELETE_TASK': {
                'access': access,
                'meta': meta,
                'method': 'delete',
                'path_conf': {
                    'path': '/hi/{id}',
                    'parameters': [{'name': 'id', 'type': 'integer'}],
                },
                'schemas': {
                    'input_body': {'body': 'delete.schema'},
                    'input_query': {'query': 'delete.schema'},
                    'output': {'output': 'delete.schema'},
                },
                'source': source,
                'examples': {},
            },
        }

    def test_render__with_examples(self):
        renderer = self.mocker.patch(
            'lily.entrypoint.renderer.SchemaRenderer')
        serialize = Mock()
        renderer.return_value.render.return_value = Mock(
            serialize=serialize, enums=[])
        serialize.side_effect = [
            {'output': 'schema'},
            {'query': 'schema'},
            {'body': 'schema'},
        ]

        meta = Meta(
            title='hi',
            description='ho',
            domain=Domain(id='h', name='hh'))
        access = Access(access_list=['EVERYONE'], is_private=True)
        source = Source(fn)
        self.mocker.patch.object(BaseRenderer, 'render').return_value = {
            'READ_CARD': {
                'method': 'get',
                'path_conf': {
                    'path': '/hi',
                    'pattern': '/hi',
                    'parameters': [],
                },
                'meta': meta,
                'access': access,
                'input': Input(query_parser=Mock(), body_parser=Mock()),
                'output': Output(serializer=Mock()),
                'source': source,
            }
        }

        with open(self.examples_filepath, 'w') as f:
            f.write(json.dumps({
                'READ_CARD': {
                    '200 (OK)': {
                        'request': {
                            'path': '/hi',
                            'parameters': {},
                        },
                    },
                },
            }))

        assert CommandsRenderer().render() == {
            '@enums': [],
            'READ_CARD': {
                'access': access,
                'meta': meta,
                'method': 'get',
                'path_conf': {
                    'parameters': [],
                    'path': '/hi',
                },
                'schemas': {
                    'input_body': {'body': 'schema'},
                    'input_query': {'query': 'schema'},
                    'output': {'output': 'schema'},
                },
                'source': source,
                'examples': {
                    '200 (OK)': {
                        'request': {
                            'path': '/hi',
                            'parameters': {},
                        },
                    },
                },
            }
        }

    def test_render__no_commands(self):
        self.mocker.patch.object(BaseRenderer, 'render').return_value = {}

        assert CommandsRenderer().render() == {'@enums': []}

    #
    # get_examples
    #
    def test_get_examples(self):
        renderer = CommandsRenderer()
        renderer.examples = {
            'COMMAND_A': {
                '200 (OK)': {
                    'request': {
                        'path': '/a/a/'
                    },
                    'response': {
                        'content': 'hi',
                    }
                },
            },
            'COMMAND_B': {
                'other': 'stuff',
            }
        }

        assert renderer.get_examples('COMMAND_A', r'/a/a/') == {
            '200 (OK)': {
                'request': {
                    'path': '/a/a/',
                    'parameters': {},
                },
                'response': {
                    'content': 'hi',
                },
            }
        }

    def test_get_examples__no_examples(self):
        renderer = CommandsRenderer()
        renderer.examples = {
            'COMMAND_A': {'hi': 'there'},
        }

        assert renderer.get_examples('COMMAND_B', r'') == {}

    def test_get_examples__with_pamaterized_paths(self):
        renderer = CommandsRenderer()
        renderer.examples = {
            'COMMAND_A': {
                '200 (OK)': {
                    'request': {
                        'path': '/a/12/b/there'
                    },
                    'response': {
                        'content': 'hi',
                    }
                },
            },
            'COMMAND_B': {
                'other': 'stuff',
            }
        }

        assert renderer.get_examples(
            'COMMAND_A',
            r'/a/(?P<id>\d+)/b/(?P<name>\w+)'
        ) == {
            '200 (OK)': {
                'request': {
                    'path': '/a/12/b/there',
                    'parameters': {
                        'id': '12',
                        'name': 'there',
                    },
                },
                'response': {
                    'content': 'hi',
                },
            }
        }




import os
import re

import click
from lily_assistant.config import Config


class Copier:

    def __init__(self):
        self.root_dir = os.getcwd()

    def copy(self, src_dir):

        self.copy_makefile(src_dir)

    def copy_makefile(self, src_dir):

        config = Config()

        with open(self.base_makefile_path, 'r') as makefile:
            content = makefile.read()
            content = re.sub(r'{%\s*SRC_DIR\s*%}', src_dir, content)
            content = re.sub(r'{%\s*VERSION\s*%}', config.version, content)

        if not os.path.exists(os.path.join(self.root_dir, '.lily')):
            os.mkdir(os.path.join(self.root_dir, '.lily'))

        makefile_path = os.path.join(
            self.root_dir, '.lily', 'lily.makefile')

        with open(makefile_path, 'w') as f:
            f.write(content)

        click.secho(
            'copied lily makefile to {makefile_path}'.format(
                makefile_path=makefile_path),
            fg='blue')

    @property
    def base_makefile_path(self):

        return os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'base.makefile')




from copy import copy, deepcopy
import re

from django.urls import URLResolver, URLPattern

from lily.base.events import EventFactory
from lily.base.command import HTTPCommands


class BaseRenderer(EventFactory):

    def __init__(self, urlpatterns):
        self.urlpatterns = urlpatterns

    def render(self):
        views_index = self.crawl_views(self.urlpatterns)

        self.validate_index(views_index)

        commands_index = {}
        for path, view_conf in views_index.items():
            view = views_index[path]['callback'].view_class

            for method in ['post', 'get', 'put', 'delete']:
                try:
                    fn = getattr(view, method)

                except AttributeError:
                    pass

                else:
                    try:
                        command_name = fn.command_conf['name']

                    except AttributeError:
                        raise self.BrokenRequest(
                            'NOT_LILY_COMPATIBLE_VIEW_DETECTED',
                            data={
                                'name': views_index[path]['name'],
                            })

                    if command_name in commands_index:
                        existing = commands_index[command_name]

                        raise self.BrokenRequest(
                            'DUPLICATED_COMMAND_DETECTED',
                            data={
                                'command_name': command_name,
                                'existing_command': {
                                    'path': existing['path_conf']['path'],
                                    'method': existing['method'].upper(),
                                },
                                'duplicate_command': {
                                    'path': path,
                                    'method': method.upper(),
                                },
                            })

                    commands_index[command_name] = {
                        'method': method.upper(),
                        'path_conf': deepcopy(views_index[path]['path_conf']),
                    }
                    commands_index[command_name].update(fn.command_conf)

        return commands_index

    def crawl_views(self, patterns, path_patterns=None):
        views_index = {}

        for pattern in patterns:
            inner_path_patterns = copy(path_patterns)
            if inner_path_patterns:
                inner_path_patterns.append(pattern.pattern.regex.pattern)

            else:
                inner_path_patterns = [pattern.pattern.regex.pattern]

            if isinstance(pattern, URLResolver):
                views_index.update(
                    self.crawl_views(
                        pattern.url_patterns, inner_path_patterns))

            elif isinstance(pattern, URLPattern):
                # -- allow only the valid lily commands to be considered
                is_lily_commands = (
                    hasattr(pattern.callback, '__self__') and
                    isinstance(pattern.callback.__self__, HTTPCommands)
                ) or (
                    hasattr(pattern.callback, 'view_class') and
                    issubclass(pattern.callback.view_class, HTTPCommands)
                )

                if not is_lily_commands:
                    continue

                view_name = pattern.callback.__name__

                ignore_views = [
                    'serve',
                    'add_view',
                    'change_view',
                    'changelist_view',
                    'history_view',
                    'delete_view',
                    'RedirectView',
                ]

                if view_name not in ignore_views:
                    path_conf = self.url_pattern_to_dict(inner_path_patterns)
                    views_index[path_conf['path']] = {
                        'name': view_name,
                        'path_conf': path_conf,
                        'path_patterns': inner_path_patterns,
                        'callback': pattern.callback,
                    }

        return views_index

    def url_pattern_to_dict(self, patterns):

        pattern = ''.join(patterns)

        # -- remove starting `^`
        pattern = re.sub(r'^\^', '/', pattern)
        # -- remove ending `$`
        pattern = re.sub(r'\$$', '', pattern)
        # -- replace middle `/^` with optional `$` (`/$^`)
        pattern = re.sub(r'\/?\$?\^', '/', pattern)
        # -- replace double `//`
        pattern = re.sub(r'\/\/', '/', pattern)

        # -- params
        parameter_types = [
            {'pattern': '\\w+', 'type': 'string'},
            {'pattern': '\\d+', 'type': 'integer'},
        ]

        parameters = []
        path = pattern
        while True:
            param_pattern = re.compile(
                r'\(\?P\<(?P<param_name>\w+)\>(?P<param_pattern>[^\)]+)\)')

            match = param_pattern.search(path)
            if match:
                span = match.span()
                param_name = match.groupdict()['param_name']
                param_pattern = match.groupdict()['param_pattern']
                param_type = 'string'
                for parameter_type_def in parameter_types:
                    if parameter_type_def['pattern'] == param_pattern:
                        param_type = parameter_type_def['type']
                        break

                path = path.replace(
                    path[span[0]:span[1]],
                    '{{{name}}}'.format(name=param_name))

                parameters.append({
                    'name': param_name,
                    'type': param_type,
                    # FIXME: add description to the parameters
                    'description': '',
                })

            else:
                break

        return {
            'path': path,
            'pattern': pattern,
            'parameters': parameters
        }

    def validate_index(self, views_index):

        view_paths_count = {}
        for view_conf in views_index.values():
            view_name = view_conf['callback'].view_class.__name__

            view_paths_count.setdefault(view_name, 0)
            view_paths_count[view_name] += 1

        duplicates = [
            view_name
            for view_name, view_path_count in view_paths_count.items()
            if view_path_count > 1]

        if duplicates:
            raise self.ServerError(
                'VIEWS_BELONGING_TO_MULTIPLE_PATHS_DETECTED',
                data={'duplicates': duplicates})




import random
from faker import Faker
from lily import Meta, Domain, Access, Source


faker = Faker()


class EntityFactory:

    def command(self, is_private=None, domain_id=None):

        def fn():
            pass

        if is_private is None:
            is_private = random.choice([True, False])

        if domain_id is None:
            domain_id = faker.word()

        command = {
            'method': random.choice(['GET', 'POST', 'PUT', 'DELETE']),
            'path_conf': {'path': 'conf'},
            'meta': Meta(
                title=faker.sentence(),
                description=faker.sentence(),
                domain=Domain(id=domain_id, name='domain')),
            'access': Access(
                is_private=is_private,
                access_list=['ANY']),
            'source': Source(fn),
            'schemas': {'some': 'schemas'},
            'examples': {'some': 'examples'},
        }

        return command




from django.test import TestCase

from lily.base import parsers


class RequestGet(dict):

    def getlist(self, name, alternative=None):
        return self.get(name, alternative)

    # -- RequestGet should behave as request.GET
    def __getitem__(self, name):
        return super(RequestGet, self).__getitem__(name)[0]


class QueryParserTestCase(TestCase):

    def test_init__boolean_field(self):
        class QueryParser(parsers.Parser):
            is_ready = parsers.BooleanField()

        cases = [
            (['true'], True),
            (['True'], True),
            (['t'], True),
            (['T'], True),
            (['false'], False),
            (['False'], False),
            (['f'], False),
            (['F'], False),
        ]
        for input_value, expected in cases:
            parser = QueryParser.as_query_parser(
                data=RequestGet(is_ready=input_value))

            assert parser.is_valid() is True
            assert parser.data == {'is_ready': expected}

    def test_init__list_boolean_field(self):
        class QueryParser(parsers.Parser):

            is_ready = parsers.ListField(child=parsers.BooleanField())

        cases = [
            ([], []),
            (['true'], [True]),
            (['True', 'False'], [True, False]),
            (['t', 'f', 'true'], [True, False, True]),
            (['T', 'F', 'f'], [True, False, False]),
        ]
        for input_value, expected in cases:
            parser = QueryParser.as_query_parser(
                data=RequestGet(is_ready=input_value))

            assert parser.is_valid() is True
            assert parser.data == {'is_ready': expected}

    def test_init__number_fields(self):
        class QueryParser(parsers.Parser):
            price = parsers.IntegerField(default=12)
            value = parsers.FloatField(default=18.5)

        cases = [
            (([1], [2.3]), (1, 2.3)),
            ((['1'], ['2.3']), (1, 2.3)),
            ((['11'], ['122']), (11, 122.0)),
            (([], []), (12, 18.5)),
        ]
        for input_value, expected in cases:
            price, value = input_value
            expected_price, expected_value = expected
            parser = QueryParser.as_query_parser(
                data=RequestGet(price=price, value=value))

            assert parser.is_valid() is True
            assert parser.data == {
                'price': expected_price, 'value': expected_value}

    def test_init__list_number_fields(self):
        class QueryParser(parsers.Parser):
            price = parsers.ListField(child=parsers.IntegerField(default=12))
            value = parsers.ListField(child=parsers.FloatField(default=18.5))

        cases = [
            (([1], [2.3]), ([1], [2.3])),
            ((['1', '-9'], ['2.3', 78, '6.78']), ([1, -9], [2.3, 78.0, 6.78])),
        ]
        for input_value, expected in cases:
            price, value = input_value
            expected_price, expected_value = expected
            parser = QueryParser.as_query_parser(
                data=RequestGet(price=price, value=value))

            assert parser.is_valid() is True
            assert parser.data == {
                'price': expected_price, 'value': expected_value}

    def test_init__data_is_optional(self):
        class QueryParser(parsers.Parser):
            is_ready = parsers.BooleanField()

        # -- no error is raised
        QueryParser.as_query_parser()


class PageQueryParstCase(TestCase):

    def test_parse(self):

        parser = parsers.PageParser.as_query_parser(
            data={'offset': '11', 'limit': 456})

        assert parser.is_valid() is True
        assert parser.data == {'offset': 11, 'limit': 456}

    def test_parse__defaults(self):

        parser = parsers.PageParser.as_query_parser(data={})

        assert parser.is_valid() is True
        assert parser.data == {'offset': 0, 'limit': 100}

    def test_parse__invalid(self):

        parser = parsers.PageParser.as_query_parser(
            data={'offset': 'HEY', 'limit': 'X'})

        assert parser.is_valid() is False
        assert parser.errors == {
            'limit': ['A valid integer is required.'],
            'offset': ['A valid integer is required.'],
        }


class FullTextSearchParserTestCase(TestCase):

    def test_parse(self):

        parser = parsers.FullTextSearchParser.as_query_parser(
            data={'query': 'hello'})

        assert parser.is_valid() is True
        assert parser.data == {'query': 'hello'}

    def test_parse__defaults(self):

        parser = parsers.FullTextSearchParser.as_query_parser(data={})

        assert parser.is_valid() is True
        assert parser.data == {'query': None}




from enum import Enum

from django.test import TestCase
from django.db import models
from django.db.utils import DataError
from django_fake_model import models as fake_models
from django.core.exceptions import ValidationError
import pytest

from lily.base.models import (
    array,
    boolean,
    const,
    enum,
    EnumChoiceField,
    ImmutableModel,
    JSONSchemaField,
    multischema,
    number,
    object,
    one_of,
    string,
    ValidatingModel,
)


class SchemaHelpersTestCase(TestCase):

    #
    # ENUM
    #
    def test_enum__from_list(self):

        assert enum('Jack', 'Alice') == {
            'enum': ['Jack', 'Alice'],
            'enum_name': None,
            'type': 'string',
        }
        assert enum(1, 11) == {
            'enum': [1, 11],
            'enum_name': None,
            'type': 'integer',
        }

    def test_enum__from_enum(self):

        class Names(Enum):
            JACK = 'Jack'

            ALICE = 'Alice'

        assert enum(Names) == {
            'enum': ['Jack', 'Alice'],
            'enum_name': 'Names',
            'type': 'string',
        }

        class Ages(Enum):
            kid = 10

            adult = 21

        assert enum(Ages) == {
            'enum': [10, 21],
            'enum_name': 'Ages',
            'type': 'integer',
        }


class ImmutableEntity(fake_models.FakeModel, ImmutableModel):

    name = models.CharField(max_length=10)


@ImmutableEntity.fake_me
class ImmutableModelTestCase(TestCase):

    def test_save_not_allowed_for_existing_instance(self):

        e = ImmutableEntity(name='john')
        e.save()

        with pytest.raises(ImmutableModel.ModelIsImmutableError):
            e.name = 'mark'
            e.save()

    def test_create_allowed(self):

        e = ImmutableEntity.objects.create(name='john')

        assert e.name == 'john'

    def test_validates_on_save(self):

        with pytest.raises(ValidationError) as e:
            ImmutableEntity(name=11 * 'a').save()

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }


class JSONModel(fake_models.FakeModel):

    answers = JSONSchemaField(schema=array(
        object(
            answer=string(),
            correct=boolean(),
            required=['answer', 'correct'])))


class JSONMultiSchemaModel(fake_models.FakeModel):

    entity = JSONSchemaField(schema=multischema(
        {
            'PERSON': object(
                type=const('PERSON'),
                name=string(),
                required=['type', 'name']),
            'ANIMAL': object(
                type=const('ANIMAL'),
                age=number(),
                required=['type', 'age'])
        },
        by_field='type'))


class JSONArrayMultiSchemaModel(fake_models.FakeModel):

    class EntityType(Enum):

        WORKER = 'WORKER'

        FREE = 'FREE'

    entities = JSONSchemaField(schema=array(
        one_of(
            object(
                name=string(),
                type=enum(EntityType, const=EntityType.WORKER.value),
                required=['type', 'name']),
            object(
                age=number(),
                type=enum(EntityType, const=EntityType.FREE.value),
                required=['type', 'age']),
            by_field='type')))


@JSONModel.fake_me
@JSONMultiSchemaModel.fake_me
@JSONArrayMultiSchemaModel.fake_me
class JSONSchemaFieldTestCase(TestCase):

    #
    # JSONModel
    #
    def test_valid_model(self):

        m = JSONModel(answers=[
            {'answer': 'hello', 'correct': True},
        ])
        m.save()

        assert m.answers == [
            {'answer': 'hello', 'correct': True},
        ]

    def test_invalid_model__empty_array(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers=[]).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == 'This field cannot be blank.'

    def test_invalid_model__not_array(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers={'hi': 'there'}).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == (
            "JSON did not validate. PATH: '.' REASON: {'hi': 'there'} "
            "is not of type 'array'"
        )

    def test_invalid_model__invalid_items__wrong_fields(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers=[
                {'answer': 'hi', 'correct': True},
                {'NOT_answer': 'hi', 'WHAT_correct': True},
            ]).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == (
            "JSON did not validate. "
            "PATH: '1' REASON: 'answer' is a required property"
        )

    #
    # JSONMultiSchemaModel
    #
    def test_valid_model__multischema(self):

        # -- schema PERSON
        m = JSONMultiSchemaModel(entity={
            'type': 'PERSON',
            'name': 'Roger',
        })
        m.save()

        assert m.entity == {
            'type': 'PERSON',
            'name': 'Roger',
        }

        # -- schema ANIMAL
        m = JSONMultiSchemaModel(entity={
            'type': 'ANIMAL',
            'age': 45,
        })
        m.save()

        assert m.entity == {
            'type': 'ANIMAL',
            'age': 45,
        }

    def test_invalid_model__multischema__missing_schema(self):

        with pytest.raises(ValidationError) as e:
            JSONMultiSchemaModel(entity={
                'type': 'ALIEN',
                'age': 45,
            }).clean_fields()

        error = e.value.error_dict['entity'][0]
        assert error.message == (
            "JSON did not validate. PATH: 'type' REASON: 'ALIEN' is "
            "not one of ['ANIMAL', 'PERSON']"
        )

    def test_invalid_model__multischema(self):

        # -- schema PERSON
        with pytest.raises(ValidationError) as e:
            JSONMultiSchemaModel(entity={
                'type': 'PERSON',
            }).clean_fields()

        error = e.value.error_dict['entity'][0]
        assert error.message == (
            "JSON did not validate. PATH: '.' REASON: 'name' is a "
            "required property"
        )

        # -- schema ANIMAL
        with pytest.raises(ValidationError) as e:
            JSONMultiSchemaModel(entity={
                'type': 'ANIMAL',
                'age': 'WHAT',
            }).clean_fields()

        error = e.value.error_dict['entity'][0]
        assert error.message == (
            "JSON did not validate. PATH: 'age' REASON: 'WHAT' is not of "
            "type 'number'"
        )

    #
    # JSONArrayMultiSchemaModel
    #
    def test_valid_model__array_multischema(self):

        # -- schema PERSON
        m = JSONArrayMultiSchemaModel(
            entities=[{'type': 'WORKER', 'name': 'Roger'}])
        m.save()

        assert m.entities == [{'type': 'WORKER', 'name': 'Roger'}]

        # -- schema ANIMAL
        m = JSONArrayMultiSchemaModel(
            entities=[{'type': 'FREE', 'age': 45}])
        m.save()

        assert m.entities == [{'type': 'FREE', 'age': 45}]

    def test_invalid_model__array_multischema__missing_schema(self):

        with pytest.raises(ValidationError) as e:
            JSONArrayMultiSchemaModel(
                entities=[{'type': 'BOSS', 'name': 'Roger'}]
            ).clean_fields()

        error = e.value.error_dict['entities'][0]
        assert error.message == (
            "JSON did not validate. PATH: 'type' REASON: 'BOSS' is "
            "not one of ['FREE', 'WORKER']"
        )

    def test_invalid_model__array_multischema(self):

        # -- schema WORKER
        with pytest.raises(ValidationError) as e:
            JSONArrayMultiSchemaModel(
                entities=[{'type': 'WORKER'}]
            ).clean_fields()

        error = e.value.error_dict['entities'][0]
        assert error.message == (
            "JSON did not validate. PATH: '.' REASON: 'name' is a "
            "required property"
        )

        # -- schema FREE
        with pytest.raises(ValidationError) as e:
            JSONArrayMultiSchemaModel(
                entities=[{'type': 'FREE', 'age': 'what'}]
            ).clean_fields()

        error = e.value.error_dict['entities'][0]
        assert error.message == (
            "JSON did not validate. PATH: 'age' REASON: 'what' is not of "
            "type 'number'"
        )


class EnumModel(fake_models.FakeModel):

    answer = EnumChoiceField(
        enum_name='super_enum',
        max_length=8,
        choices=[('A', 'A'), ('B', 'B')])

    class Names(Enum):

        ALICE = 'ALICE'

        JACK = 'JACK'

    name = EnumChoiceField(enum=Names, max_length=8)


@EnumModel.fake_me
class EnumChoiceFieldTestCase(TestCase):

    def test_model(self):

        m = EnumModel(answer='A', name='JACK')
        m.save()

        assert m.answer == 'A'
        assert m.name == 'JACK'

    def test_extra_field_attributes(self):

        for field in EnumModel._meta.fields:
            if field.name == 'answer':
                assert field.enum_name == 'super_enum'

            if field.name == 'name':
                assert field.enum_name == 'Names'
                assert field.choices == [('ALICE', 'ALICE'), ('JACK', 'JACK')]


class ValidatingEntity(fake_models.FakeModel, ValidatingModel):

    name = models.CharField(max_length=10)


class ValidatingEntityWithJSON(fake_models.FakeModel, ValidatingModel):

    tasks = JSONSchemaField(
        schema=array(
            object(
                type=const('TASK'),
                name=string())))

    def clean(self):
        if len(set([t['name'] for t in self.tasks])) < 2:
            raise ValidationError('at least 2 unique task needed')


@ValidatingEntity.fake_me
@ValidatingEntityWithJSON.fake_me
class ValidatingModelTestCase(TestCase):

    #
    # ValidatingEntity
    #
    def test_validates_on_save(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntity(name=11 * 'a').save()

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }

    def test_validates_on_create(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntity.objects.create(name=11 * 'a')

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }

    def test_validates_on_bulk_create(self):

        with pytest.raises(DataError):
            ValidatingEntity.objects.bulk_create([
                ValidatingEntity(name=11 * 'a'),
            ])

    #
    # ValidatingEntityWithJSON
    #
    def test_schema_validation__run_before_clean(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntityWithJSON.objects.create(tasks={'what': 'hey'})

        assert e.value.message_dict == {
            'tasks': [
                "JSON did not validate. PATH: '.' REASON: {'what': 'hey'} "
                "is not of type 'array'",
            ],
        }

    def test_schema_validation_when_ok_clean_is_called(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntityWithJSON.objects.create(tasks=[
                {'name': 'hey', 'type': 'TASK'},
            ])

        assert e.value.message_dict == {
            '__all__': ['at least 2 unique task needed'],
        }

    def test_schema_validation_when_const_is_broken(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntityWithJSON.objects.create(tasks=[
                {'name': 'hey', 'type': 'TASKY'},
            ])

        assert e.value.message_dict == {
            'tasks': [
                "JSON did not validate. PATH: '0.type' REASON: 'TASKY' "
                "does not match '^TASK$'",
            ],
        }




from unittest import TestCase
from unittest.mock import Mock

from django.urls import re_path, include
import pytest

from lily.entrypoint.base import BaseRenderer
from lily.base.command import HTTPCommands
from lily.base.events import EventFactory


class BaseRendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    #
    # RENDER
    #
    def test_render__simple_flat_url_patterns(self):

        class HiCommands(HTTPCommands):
            def post(self):
                pass

            post.command_conf = {
                'name': 'Hi',
                'some': 'hi.conf',
            }

        class HelloCommands(HTTPCommands):
            def delete(self):
                pass

            delete.command_conf = {
                'name': 'Hello',
                'some': 'hello.conf',
            }

        hi_cmds, hello_cmds = HiCommands.as_view(), HelloCommands.as_view()

        renderer = BaseRenderer([
            re_path(r'^hi/there$', hi_cmds, name='hi.there'),
            re_path(r'^hiyo', hello_cmds, name='hiyo'),
        ])

        assert renderer.render() == {
            'Hi': {
                'method': 'POST',
                'path_conf': {
                    'path': '/hi/there',
                    'pattern': r'/hi/there',
                    'parameters': [],
                },
                'name': 'Hi',
                'some': 'hi.conf',
            },
            'Hello': {
                'method': 'DELETE',
                'path_conf': {
                    'path': '/hiyo',
                    'pattern': '/hiyo',
                    'parameters': [],
                },
                'name': 'Hello',
                'some': 'hello.conf',
            },
        }

    def test_render__same_path_many_commands(self):

        class HiCommands(HTTPCommands):
            def post(self):
                pass

            post.command_conf = {
                'name': 'Hi',
                'some': 'hi.conf',
            }

            def delete(self):
                pass

            delete.command_conf = {
                'name': 'Hello',
                'some': 'hello.conf',
            }

        hi_cmds = HiCommands.as_view()

        renderer = BaseRenderer([
            re_path(r'^hi/there$', hi_cmds, name='hi.there'),
        ])

        assert renderer.render() == {
            'Hi': {
                'method': 'POST',
                'path_conf': {
                    'path': '/hi/there',
                    'pattern': r'/hi/there',
                    'parameters': [],
                },
                'name': 'Hi',
                'some': 'hi.conf',
            },
            'Hello': {
                'method': 'DELETE',
                'path_conf': {
                    'path': '/hi/there',
                    'pattern': r'/hi/there',
                    'parameters': [],
                },
                'name': 'Hello',
                'some': 'hello.conf',
            },
        }

    def test_render__command_belongs_to_many_paths_error(self):

        class HiCommands(HTTPCommands):
            def post(self):
                pass

            post.command_conf = {
                'name': 'Hi',
                'some': 'hi.conf',
            }

        hi_cmds = HiCommands.as_view()

        renderer = BaseRenderer([
            re_path(r'^hi/there$', hi_cmds, name='hi.there'),
            re_path(r'^hello/there$', hi_cmds, name='hello.there'),
        ])

        with pytest.raises(EventFactory.ServerError) as e:
            renderer.render()

        assert e.value.event == 'VIEWS_BELONGING_TO_MULTIPLE_PATHS_DETECTED'
        assert e.value.data == {
            '@event': 'VIEWS_BELONGING_TO_MULTIPLE_PATHS_DETECTED',
            '@type': 'error',
            'duplicates': ['HiCommands'],
        }

    def test_render__not_lily_compatible_cmds(self):

        class HiCommands(HTTPCommands):
            def post(self):
                pass

        renderer = BaseRenderer([
            re_path(r'^hi/there$', HiCommands.as_view(), name='hi.there'),
        ])

        with pytest.raises(EventFactory.BrokenRequest) as e:
            renderer.render()

        assert e.value.data == {
            '@event': 'NOT_LILY_COMPATIBLE_VIEW_DETECTED',
            '@type': 'error',
            'name': 'HiCommands',
        }

    def test_render__duplicated_command(self):

        class HiCommands(HTTPCommands):

            def post(self):
                pass

            post.command_conf = {
                'name': 'HI',
                'some': 'hi.conf',
            }

        class HelloCommands(HTTPCommands):

            def delete(self):
                pass

            delete.command_conf = {
                'name': 'HI',
                'some': 'hello.conf',
            }

        renderer = BaseRenderer([
            re_path(r'^root/hi$', HiCommands.as_view(), name='hi'),

            re_path(r'^root/hello$', HelloCommands.as_view(), name='hello'),
        ])

        with pytest.raises(EventFactory.BrokenRequest) as e:
            renderer.render()

        assert e.value.data == {
            '@event': 'DUPLICATED_COMMAND_DETECTED',
            '@type': 'error',
            'command_name': 'HI',
            'duplicate_command': {
                'path': '/root/hello',
                'method': 'DELETE',
            },
            'existing_command': {
                'path': '/root/hi',
                'method': 'POST',
            },
        }

    #
    # CRAWL_VIEWS
    #
    def test_crawl_commands__deep_url_patterns(self):

        class HiCommands(HTTPCommands):
            def post(self):
                pass

            post.command_conf = {
                'name': 'Hi',
                'conf': 'hi',
            }

        class WatCommands(HTTPCommands):
            def get(self):
                pass

            get.command_conf = {
                'name': 'Wat',
                'conf.wat': 'wat',
            }

        class YoCommands(HTTPCommands):
            def delete(self):
                pass

            delete.command_conf = {
                'name': 'Yo',
                'conf': 'yo',
            }

        hi_cmds, yo_cmds, wat_cmds = (
            HiCommands.as_view(), YoCommands.as_view(), WatCommands.as_view())

        renderer = BaseRenderer([
            re_path(r'^payment/$', include([
                re_path(r'^hi/there$', hi_cmds, 'hi.there'),
                re_path(r'^now/(?P<when>\d+)/$', include([
                    re_path(r'^yes/$', yo_cmds, 'yo'),
                    re_path(r'^wat/$', wat_cmds, 'wat'),
                ])),
            ])),
        ])

        assert renderer.render() == {
            'Hi': {
                'name': 'Hi',
                'method': 'POST',
                'path_conf': {
                    'path': '/payment/hi/there',
                    'pattern': r'/payment/hi/there',
                    'parameters': [],
                },
                'conf': 'hi',
            },
            'Yo': {
                'name': 'Yo',
                'method': 'DELETE',
                'path_conf': {
                    'path': '/payment/now/{when}/yes/',
                    'pattern': r'/payment/now/(?P<when>\d+)/yes/',
                    'parameters': [
                        {
                            'name': 'when',
                            'description': '',
                            'type': 'integer',
                        }
                    ],
                },
                'conf': 'yo',
            },
            'Wat': {
                'name': 'Wat',
                'method': 'GET',
                'path_conf': {
                    'path': '/payment/now/{when}/wat/',
                    'pattern': r'/payment/now/(?P<when>\d+)/wat/',
                    'parameters': [
                        {
                            'name': 'when',
                            'description': '',
                            'type': 'integer',
                        }
                    ],
                },
                'conf.wat': 'wat',
            },
        }

    #
    # url_pattern_to_conf
    #
    def test_url_pattern_to_conf__returns_path_conf__no_params(self):

        renderer = BaseRenderer(Mock())

        assert renderer.url_pattern_to_dict(
            ['^payments/', '^subscriptions/register/$']
        ) == {
            'path': '/payments/subscriptions/register/',
            'pattern': '/payments/subscriptions/register/',
            'parameters': []
        }

        # -- with extra $
        assert renderer.url_pattern_to_dict(
            ['^payments/$', '^subscriptions/register/$', '^/now$']
        ) == {
            'path': '/payments/subscriptions/register/now',
            'pattern': '/payments/subscriptions/register/now',
            'parameters': []
        }

    def test_url_pattern_to_conf__returns_path_conf__int_param(self):

        renderer = BaseRenderer(Mock())

        assert renderer.url_pattern_to_dict(
            ['^payments/', '^payment_cards/(?P<payment_card_id>\\d+)$']
        ) == {
            'path': '/payments/payment_cards/{payment_card_id}',
            'pattern': '/payments/payment_cards/(?P<payment_card_id>\\d+)',
            'parameters': [
                {
                    'name': 'payment_card_id',
                    'description': '',
                    'type': 'integer',
                }
            ]
        }

    def test_url_pattern_to_conf__returns_path_conf__string_param(self):

        renderer = BaseRenderer(Mock())

        assert renderer.url_pattern_to_dict([
            '^payments/',
            '^payment_cards/(?P<card_id>\\w+)/mark_as_default/$'
        ]) == {
            'path': '/payments/payment_cards/{card_id}/mark_as_default/',
            'pattern': (
                '/payments/payment_cards/(?P<card_id>\\w+)/mark_as_default/'),
            'parameters': [
                {
                    'name': 'card_id',
                    'description': '',
                    'type': 'string',
                }
            ]
        }

    def test_url_pattern_to_conf__returns_path_conf__complex_string_param(
            self):

        renderer = BaseRenderer(Mock())

        result = renderer.url_pattern_to_dict([
            '^/mediaitems/uploads/sign/'
            '(?P<upload_file_properties>[\\w\\=\\+\\/]+)',
        ])
        assert result == {
            'path': '/mediaitems/uploads/sign/{upload_file_properties}',
            'pattern': (
                '/mediaitems/uploads/sign/'
                '(?P<upload_file_properties>[\\w\\=\\+\\/]+)'),
            'parameters': [
                {
                    'name': 'upload_file_properties',
                    'description': '',
                    'type': 'string',
                }
            ]
        }

    def test_url_pattern_to_conf__returns_path_conf__many_params(self):

        renderer = BaseRenderer(Mock())

        assert renderer.url_pattern_to_dict([
            '^cards/',
            '^(?P<card_id>\\w+)$',
            '^/with/$',
            '^(?P<messageId>\\d+)/$'
        ]) == {
            'path': '/cards/{card_id}/with/{messageId}/',
            'pattern': (
                '/cards/(?P<card_id>\\w+)/with/(?P<messageId>\\d+)/'),
            'parameters': [
                {
                    'name': 'card_id',
                    'description': '',
                    'type': 'string',
                },
                {
                    'name': 'messageId',
                    'description': '',
                    'type': 'integer',
                },
            ]
        }




import os
import sys
from importlib import import_module

import djclick as click
from lily_assistant.config import Config

from lily.conf import settings
from ...renderers.markdown.renderer import MarkdownRenderer


@click.command()
def command():
    """Render Markdown Specification for all registered Commands."""

    # -- make sure that the main directory is also visible during
    # -- the search of all url patterns
    sys.path.insert(0, os.getcwd())

    urlpatterns = import_module(settings.ROOT_URLCONF).urlpatterns

    with open(os.path.join(Config.get_lily_path(), 'API.md'), 'w') as f:
        f.write(MarkdownRenderer(urlpatterns).render())

    click.secho(
        'Successfully rendered Markdown Specification for all '
        'registered Commands',
        fg='green')




from importlib import import_module
import json
import os
import re
import sys

from lily.conf import settings
from lily.base.test import get_examples_filepath
from lily.base import parsers
from .base import BaseRenderer
from .schema import SchemaRenderer


class CommandsRenderer(BaseRenderer):

    def __init__(self):
        self.urlpatterns = self.get_urlpatterns()
        self.examples = self.get_all_examples()

    def get_urlpatterns(self):

        # -- make sure that the main directory is also visible during
        # -- the search of all url patterns
        sys.path.insert(0, os.getcwd())

        return import_module(settings.ROOT_URLCONF).urlpatterns

    def render(self):

        commands_index = super(CommandsRenderer, self).render()
        rendered = {}
        enums = []

        for name, conf in commands_index.items():

            path_conf = conf['path_conf']
            method = conf['method']
            access = conf['access']
            input_ = conf['input']
            output = conf['output']
            source = conf['source']
            meta = self.resolve_description(conf['meta'], source)

            configuration = {
                'method': method,
                'path_conf': path_conf,
                'meta': meta,
                'source': source,
                'access': access,
            }

            # -- EXAMPLES
            pattern = path_conf.pop('pattern')
            configuration['examples'] = self.get_examples(name, pattern)

            # -- SCHEMAS
            schemas = {}
            output_schema = SchemaRenderer(output.serializer).render()
            schemas['output'] = output_schema.serialize()
            enums.extend(output_schema.enums)

            if input_.query_parser:
                input_query_schema = (
                    SchemaRenderer(
                        input_.query_parser,
                        parser_type=parsers.ParserTypeEnum.QUERY.value
                    ).render())
                schemas['input_query'] = input_query_schema.serialize()
                enums.extend(input_query_schema.enums)

            if input_.body_parser:
                input_body_schema = (
                    SchemaRenderer(
                        input_.body_parser,
                        parser_type=parsers.ParserTypeEnum.BODY.value
                    ).render())
                schemas['input_body'] = input_body_schema.serialize()
                enums.extend(input_body_schema.enums)

            configuration['schemas'] = schemas
            rendered[name] = configuration

        rendered['@enums'] = self.unique_enums(enums)
        return rendered

    def resolve_description(self, meta, source):
        desc = meta.description

        if desc and desc.endswith('.md'):
            base = os.path.dirname(source.filepath)[1:]
            with open(os.path.join(base, desc), 'r') as f:
                meta.description = f.read().strip()

        return meta

    def unique_enums(self, enums):
        unique_enums = set(
            [json.dumps(enum, sort_keys=True) for enum in enums])

        return sorted(
            [json.loads(enum) for enum in unique_enums],
            key=lambda x: x['enum_name'])

    def get_all_examples(self):
        try:
            with open(get_examples_filepath()) as f:
                return json.loads(f.read())

        except FileNotFoundError:
            return {}

    def get_examples(self, command_name, path_pattern):

        try:
            examples = self.examples[command_name]

        except KeyError:
            return {}

        else:
            for example in examples.values():
                path = example['request']['path']
                parameters = re.compile(path_pattern).search(path)
                example['request']['parameters'] = parameters.groupdict()

            return examples




import shutil
import os
import re
from functools import partial
import tempfile

from lily_assistant.repo.repo import Repo

from lily.conf import settings


class AngularRepo(Repo):

    origin = NotImplementedError

    def __init__(self, origin):
        self.origin = origin
        self.base_path = tempfile.mkdtemp()

    def clone(self):
        super(AngularRepo, self).clone(self.base_path)

    def commit(self, version):
        self.git(f'commit -m "ADDED version {version}"')

    #
    # NPM
    #
    def build(self):
        self.npm('run build')

    def install(self):
        self.npm('install')

    def link(self):
        self.npm('link @lily/http')

    def npm(self, command):
        self.execute(f'npm {command}')

    #
    # UTILS
    #
    # FIXME: allow here passing the version from the api gateway!!!
    # to have match between client version and API version
    # upgrades of the client itself, how can we control them???
    # -- maybe by adding extra API-<version>-CLIENT-<version>
    def upgrade_version(self, config):

        with open('package.json', 'r') as p:
            conf = p.read()
            version_match = re.search(
                r'\"version\"\:\s\"(?P<version>[\d\.]+)\"', conf, re.M)
            version = version_match.groupdict()['version']
            span = version_match.span()

            api_version = config.version
            next_version = f'{api_version}-API-{version}-CLIENT'

        with open('package.json', 'w') as p:
            conf = conf.replace(
                conf[span[0]: span[1]], '"version": "{}"'.format(next_version))
            p.write(conf)

        return next_version


class PathRule:

    def __init__(self, pattern, is_directory=False):
        self.pattern = re.compile(pattern, flags=re.I)
        self.is_directory = is_directory

    def matches(self, path):

        if self.is_directory:
            match = self.pattern.search(path)

            if match:
                parent_path = path[:match.end() + 1]

                return os.path.isdir(parent_path)

            else:
                return False

        return bool(self.pattern.search(path))


class AngularHTTPRepo(AngularRepo):

    origin = settings.LILY_ANGULAR_HTTP_ORIGIN

    def __init__(self):
        self.base_path = tempfile.mkdtemp()

    def link(self):
        self.npm('link')


class TemplateRepo(Repo):

    origin = settings.LILY_ANGULAR_CLIENT_ORIGIN

    IGNORE_RULES = [
        PathRule(r'node_modules', True),
        PathRule(r'\.git', True),
    ]

    KEEP_RULES = [
        PathRule(r'.*\.ts$'),
        PathRule(r'.*\.json$'),
        PathRule(r'.*\.html$'),
        PathRule(r'.*\.md$'),
        PathRule(r'.*\.ico$'),
        PathRule(r'.*karma.conf.js$'),
        PathRule(r'.*browserslist$'),
        PathRule(r'.*\.npmignore$'),
        PathRule(r'.*\.gitkeep$'),
        PathRule(r'.*\.gitignore$'),
        PathRule(r'.*\.css$'),
        PathRule(r'.*', True),  # keep all directories not explicitly ignored
    ]

    def __init__(self):
        self.base_path = tempfile.mkdtemp()

    def clone(self):
        super(TemplateRepo, self).clone(self.base_path)

    def copy_to(self, destination, client_prefix):

        ignore = self.ignore
        copy_function = partial(self.copy, client_prefix)

        def copytree(source, destination):
            """Replace `shutil.copytree` with overwrites."""

            if os.path.isdir(source):

                if not os.path.isdir(destination):
                    os.makedirs(destination)

                names = os.listdir(source)
                ignored = ignore(source, names)

                for name in names:
                    if name not in ignored:
                        copytree(
                            os.path.join(source, name),
                            os.path.join(destination, name))

            else:
                copy_function(source, destination)

        def rmtree(destination):

            for name in os.listdir(destination):
                remove = True
                path = os.path.join(destination, name)

                # -- ignore folders which should not be removed
                for ignore_rule in self.IGNORE_RULES:
                    if ignore_rule.matches(path):
                        remove = False
                        break

                if remove:
                    if os.path.isdir(path):
                        shutil.rmtree(path)

                    else:
                        os.remove(path)

        rmtree(destination)
        copytree(self.base_path, destination)

    def ignore(self, source, names):

        ignore_names = set([])
        for name in names:
            path = os.path.join(source, name)

            for ignore_rule in self.IGNORE_RULES:
                if ignore_rule.matches(path):
                    ignore_names.add(name)
                    break

            keep = False
            for keep_rule in self.KEEP_RULES:
                if keep_rule.matches(path):
                    keep = True
                    break

            if not keep:
                ignore_names.add(name)

        return list(ignore_names)

    def copy(self, client_prefix, source, destination):

        try:
            with open(source, 'r') as f:
                content = f.read()
                content = re.sub('__CLIENT_PREFIX__', client_prefix, content)

            with open(destination, 'w') as f:
                f.write(content)

        # -- binary data copying
        except UnicodeDecodeError:
            shutil.copy2(source, destination)
